@@ -18,17 +18,20 @@ use crate::cache::azure::AzureBlobCache;
 use crate::cache::conccache::ConcurrentDiskCache;
 use crate::cache::disk::DiskCache;
 #[cfg(feature = "gcs")]
-use crate::cache::gcs::{GCSCache, RWMode};
+use crate::cache::gcs::GCSCache;
 #[cfg(feature = "gha")]
 use crate::cache::gha::GHACache;
 #[cfg(feature = "memcached")]
 use crate::cache::memcached::MemcachedCache;
+#[cfg(feature = "oss")]
+use crate::cache::oss::OSSCache;
 #[cfg(feature = "redis")]
 use crate::cache::redis::RedisCache;
 #[cfg(feature = "s3")]
 use crate::cache::s3::S3Cache;
 #[cfg(feature = "webdav")]
 use crate::cache::webdav::WebdavCache;
+use crate::compiler::PreprocessorCacheEntry;
 use crate::config::Config;
 #[cfg(any(
     feature = "azure",
@@ -38,11 +41,14 @@ use crate::config::Config;
     feature = "redis",
     feature = "s3",
     feature = "webdav",
+    feature = "oss",
     feature = "concurrent-cache"
 ))]
 use crate::config::{self, CacheType};
 use async_trait::async_trait;
 use fs_err as fs;
+
+use serde::{Deserialize, Serialize};
 use std::fmt;
 use std::io::{self, Cursor, Read, Seek, Write};
 use std::path::{Path, PathBuf};
@@ -99,6 +105,8 @@ pub enum Cache {
     Hit(CacheRead),
     /// Result was not found in cache.
     Miss,
+    /// Do not cache the results of the compilation.
+    None,
     /// Cache entry should be ignored, force compilation.
     Recache,
 }
@@ -108,13 +116,14 @@ impl fmt::Debug for Cache {
         match *self {
             Cache::Hit(_) => write!(f, "Cache::Hit(...)"),
             Cache::Miss => write!(f, "Cache::Miss"),
+            Cache::None => write!(f, "Cache::None"),
             Cache::Recache => write!(f, "Cache::Recache"),
         }
     }
 }
 
-/// CacheMode is used to repreent which mode we are using.
-#[derive(Debug)]
+/// CacheMode is used to represent which mode we are using.
+#[derive(Copy, Clone, Debug, PartialEq, Eq)]
 pub enum CacheMode {
     /// Only read cache from storage.
     ReadOnly,
@@ -288,7 +297,12 @@ impl CacheWrite {
         self.zip
             .start_file(name, opts)
             .context("Failed to start cache entry object")?;
-        zstd::stream::copy_encode(from, &mut self.zip, 3)?;
+
+        let compression_level = std::env::var("SCCACHE_CACHE_ZSTD_LEVEL")
+            .ok()
+            .and_then(|value| value.parse::<i32>().ok())
+            .unwrap_or(3);
+        zstd::stream::copy_encode(from, &mut self.zip, compression_level)?;
         Ok(())
     }
 
@@ -364,16 +378,100 @@ pub trait Storage: Send + Sync {
 
     /// Get the maximum storage size, if applicable.
     async fn max_size(&self) -> Result<Option<u64>>;
+
+    /// Return the config for preprocessor cache mode if applicable
+    fn preprocessor_cache_mode_config(&self) -> PreprocessorCacheModeConfig {
+        // Enable by default, only in local mode
+        PreprocessorCacheModeConfig::default()
+    }
+    /// Return the preprocessor cache entry for a given preprocessor key,
+    /// if it exists.
+    /// Only applicable when using preprocessor cache mode.
+    async fn get_preprocessor_cache_entry(
+        &self,
+        _key: &str,
+    ) -> Result<Option<Box<dyn crate::lru_disk_cache::ReadSeek>>> {
+        Ok(None)
+    }
+    /// Insert a preprocessor cache entry at the given preprocessor key,
+    /// overwriting the entry if it exists.
+    /// Only applicable when using preprocessor cache mode.
+    async fn put_preprocessor_cache_entry(
+        &self,
+        _key: &str,
+        _preprocessor_cache_entry: PreprocessorCacheEntry,
+    ) -> Result<()> {
+        Ok(())
+    }
+}
+
+/// Configuration switches for preprocessor cache mode.
+#[derive(Debug, Copy, Clone, PartialEq, Eq, Serialize, Deserialize)]
+#[serde(deny_unknown_fields)]
+#[serde(default)]
+pub struct PreprocessorCacheModeConfig {
+    /// Whether to use preprocessor cache mode entirely
+    pub use_preprocessor_cache_mode: bool,
+    /// If false (default), only compare header files by hashing their contents.
+    /// If true, will use size + ctime + mtime to check whether a file has changed.
+    /// See other flags below for more control over this behavior.
+    pub file_stat_matches: bool,
+    /// If true (default), uses the ctime (file status change on UNIX,
+    /// creation time on Windows) to check that a file has/hasn't changed.
+    /// Can be useful to disable when backdating modification times
+    /// in a controlled manner.
+    pub use_ctime_for_stat: bool,
+    /// If true, ignore `__DATE__`, `__TIME__` and `__TIMESTAMP__` being present
+    /// in the source code. Will speed up preprocessor cache mode,
+    /// but can result in false positives.
+    pub ignore_time_macros: bool,
+    /// If true, preprocessor cache mode will not cache system headers, only
+    /// add them to the hash.
+    pub skip_system_headers: bool,
+    /// If true (default), will add the current working directory in the hash to
+    /// distinguish two compilations from different directories.
+    pub hash_working_directory: bool,
+}
+
+impl Default for PreprocessorCacheModeConfig {
+    fn default() -> Self {
+        Self {
+            use_preprocessor_cache_mode: false,
+            file_stat_matches: false,
+            use_ctime_for_stat: true,
+            ignore_time_macros: false,
+            skip_system_headers: false,
+            hash_working_directory: true,
+        }
+    }
+}
+
+impl PreprocessorCacheModeConfig {
+    /// Return a default [`Self`], but with the cache active.
+    pub fn activated() -> Self {
+        Self {
+            use_preprocessor_cache_mode: true,
+            ..Default::default()
+        }
+    }
 }
 
 /// Implement storage for operator.
-#[cfg(any(feature = "s3", feature = "azure", feature = "gcs", feature = "redis"))]
+#[cfg(any(
+    feature = "azure",
+    feature = "gcs",
+    feature = "gha",
+    feature = "memcached",
+    feature = "redis",
+    feature = "s3",
+    feature = "webdav",
+))]
 #[async_trait]
 impl Storage for opendal::Operator {
     async fn get(&self, key: &str) -> Result<Cache> {
         match self.read(&normalize_key(key)).await {
             Ok(res) => {
-                let hit = CacheRead::from(io::Cursor::new(res))?;
+                let hit = CacheRead::from(io::Cursor::new(res.to_bytes()))?;
                 Ok(Cache::Hit(hit))
             }
             Err(e) if e.kind() == opendal::ErrorKind::NotFound => Ok(Cache::Miss),
@@ -420,7 +518,7 @@ impl Storage for opendal::Operator {
         let can_write = match self.write(path, "Hello, World!").await {
             Ok(_) => true,
             Err(err) if err.kind() == ErrorKind::AlreadyExists => true,
-            // Toralte all other write errors because we can do read as least.
+            // Tolerate all other write errors because we can do read at least.
             Err(err) => {
                 eprintln!("storage write check failed: {err:?}");
                 false
@@ -458,8 +556,7 @@ impl Storage for opendal::Operator {
 }
 
 /// Normalize key `abcdef` into `a/b/c/abcdef`
-#[allow(dead_code)]
-fn normalize_key(key: &str) -> String {
+pub(in crate::cache) fn normalize_key(key: &str) -> String {
     format!("{}/{}/{}/{}", &key[0..1], &key[1..2], &key[2..3], &key)
 }
 
@@ -493,17 +590,12 @@ pub fn storage_from_config(
             }) => {
                 debug!("Init gcs cache with bucket {bucket}, key_prefix {key_prefix}");
 
-                let gcs_read_write_mode = match rw_mode {
-                    config::GCSCacheRWMode::ReadOnly => RWMode::ReadOnly,
-                    config::GCSCacheRWMode::ReadWrite => RWMode::ReadWrite,
-                };
-
                 let storage = GCSCache::build(
                     bucket,
                     key_prefix,
                     cred_path.as_deref(),
                     service_account.as_deref(),
-                    gcs_read_write_mode,
+                    (*rw_mode).into(),
                     credential_url.as_deref(),
                 )
                 .map_err(|err| anyhow!("create gcs cache failed: {err:?}"))?;
@@ -521,19 +613,68 @@ pub fn storage_from_config(
             #[cfg(feature = "memcached")]
             CacheType::Memcached(config::MemcachedCacheConfig {
                 ref url,
+                ref username,
+                ref password,
                 ref expiration,
+                ref key_prefix,
             }) => {
                 debug!("Init memcached cache with url {url}");
 
-                let storage = MemcachedCache::build(url, *expiration)
-                    .map_err(|err| anyhow!("create memcached cache failed: {err:?}"))?;
+                let storage = MemcachedCache::build(
+                    url,
+                    username.as_deref(),
+                    password.as_deref(),
+                    key_prefix,
+                    *expiration,
+                )
+                .map_err(|err| anyhow!("create memcached cache failed: {err:?}"))?;
                 return Ok(Arc::new(storage));
             }
             #[cfg(feature = "redis")]
-            CacheType::Redis(config::RedisCacheConfig { ref url }) => {
-                debug!("Init redis cache with url {url}");
-                let storage = RedisCache::build(url)
-                    .map_err(|err| anyhow!("create redis cache failed: {err:?}"))?;
+            CacheType::Redis(config::RedisCacheConfig {
+                ref endpoint,
+                ref cluster_endpoints,
+                ref username,
+                ref password,
+                ref db,
+                ref url,
+                ref ttl,
+                ref key_prefix,
+            }) => {
+                let storage = match (endpoint, cluster_endpoints, url) {
+                    (Some(url), None, None) => {
+                        debug!("Init redis single-node cache with url {url}");
+                        RedisCache::build_single(
+                            url,
+                            username.as_deref(),
+                            password.as_deref(),
+                            *db,
+                            key_prefix,
+                            *ttl,
+                        )
+                    }
+                    (None, Some(urls), None) => {
+                        debug!("Init redis cluster cache with urls {urls}");
+                        RedisCache::build_cluster(
+                            urls,
+                            username.as_deref(),
+                            password.as_deref(),
+                            *db,
+                            key_prefix,
+                            *ttl,
+                        )
+                    }
+                    (None, None, Some(url)) => {
+                        warn!("Init redis single-node cache from deprecated API with url {url}");
+                        if username.is_some() || password.is_some() || *db != crate::config::DEFAULT_REDIS_DB {
+                            bail!("`username`, `password` and `db` has no effect when `url` is set. Please use `endpoint` or `cluster_endpoints` for new API accessing");
+                        }
+
+                        RedisCache::build_from_url(url, key_prefix, *ttl)
+                    }
+                    _ => bail!("Only one of `endpoint`, `cluster_endpoints`, `url` must be set"),
+                }
+                .map_err(|err| anyhow!("create redis cache failed: {err:?}"))?;
                 return Ok(Arc::new(storage));
             }
             #[cfg(feature = "s3")]
@@ -542,16 +683,16 @@ pub fn storage_from_config(
                     "Init s3 cache with bucket {}, endpoint {:?}",
                     c.bucket, c.endpoint
                 );
-
-                let storage = S3Cache::build(
-                    &c.bucket,
-                    c.region.as_deref(),
-                    &c.key_prefix,
-                    c.no_credentials,
-                    c.endpoint.as_deref(),
-                    c.use_ssl,
-                )
-                .map_err(|err| anyhow!("create s3 cache failed: {err:?}"))?;
+                let storage_builder =
+                    S3Cache::new(c.bucket.clone(), c.key_prefix.clone(), c.no_credentials);
+                let storage = storage_builder
+                    .with_region(c.region.clone())
+                    .with_endpoint(c.endpoint.clone())
+                    .with_use_ssl(c.use_ssl)
+                    .with_server_side_encryption(c.server_side_encryption)
+                    .with_enable_virtual_host_style(c.enable_virtual_host_style)
+                    .build()
+                    .map_err(|err| anyhow!("create s3 cache failed: {err:?}"))?;
 
                 return Ok(Arc::new(storage));
             }
@@ -570,6 +711,24 @@ pub fn storage_from_config(
 
                 return Ok(Arc::new(storage));
             }
+            #[cfg(feature = "oss")]
+            CacheType::OSS(ref c) => {
+                debug!(
+                    "Init oss cache with bucket {}, endpoint {:?}",
+                    c.bucket, c.endpoint
+                );
+
+                let storage = OSSCache::build(
+                    &c.bucket,
+                    &c.key_prefix,
+                    c.endpoint.as_deref(),
+                    c.no_credentials,
+                )
+                .map_err(|err| anyhow!("create oss cache failed: {err:?}"))?;
+
+                return Ok(Arc::new(storage));
+            }
+            #[cfg(feature = "concurrent-cache")]
             CacheType::ConcurrentDisk(config::ConcurrentDiskCacheConfig {
                 ref dir,
                 ref size,
@@ -580,7 +739,10 @@ pub fn storage_from_config(
                     dir, size, durable_fs
                 );
                 #[cfg(feature = "concurrent-cache")]
-                match ConcurrentDiskCache::new(dir, *size, *durable_fs, pool) {
+                match ConcurrentDiskCache::new(dir, *size, *durable_fs, pool,
+                    PreprocessorCacheModeConfig::default(),
+                    CacheMode::ReadWrite
+                ) {
                     Ok(s) => {
                         trace!("Using ConcurrentDiskCache");
                         return Ok(Arc::new(s));
@@ -589,25 +751,33 @@ pub fn storage_from_config(
                 }
             }
             #[allow(unreachable_patterns)]
-            _ => bail!("cache type is not enabled"),
+            // if we build only with `cargo build --no-default-features`
+            // we only want to use sccache with a local cache (no remote storage)
+            _ => {}
         }
     }
 
     let (dir, size) = (&config.fallback_cache.dir, config.fallback_cache.size);
+    let preprocessor_cache_mode_config = config.fallback_cache.preprocessor_cache_mode;
+    let rw_mode = config.fallback_cache.rw_mode.into();
     debug!("Init disk cache with dir {:?}, size {}", dir, size);
 
     #[cfg(feature = "concurrent-cache")]
     let cache = Arc::new(
-        ConcurrentDiskCache::new(&dir, size, false, pool).expect("Failed to setup cache!"),
+        ConcurrentDiskCache::new(&dir, size, false, pool,
+            preprocessor_cache_mode_config,
+            rw_mode,
+        ).expect("Failed to setup cache!"),
     );
     #[cfg(not(feature = "concurrent-cache"))]
-    let cache = Arc::new(DiskCache::new(&dir, size, pool));
+    let cache = Arc::new(DiskCache::new(&dir, size, pool, preprocessor_cache_mode_config, rw_mode));
     Ok(cache)
 }
 
 #[cfg(test)]
 mod test {
     use super::*;
+    use crate::config::CacheModeConfig;
 
     #[test]
     fn test_normalize_key() {
@@ -616,4 +786,70 @@ mod test {
             "0/1/2/0123456789abcdef0123456789abcdef"
         );
     }
+
+    #[test]
+    fn test_read_write_mode_local() {
+        let runtime = tokio::runtime::Builder::new_current_thread()
+            .enable_all()
+            .worker_threads(1)
+            .build()
+            .unwrap();
+
+        // Use disk cache.
+        let mut config = Config {
+            cache: None,
+            ..Default::default()
+        };
+
+        let tempdir = tempfile::Builder::new()
+            .prefix("sccache_test_rust_cargo")
+            .tempdir()
+            .context("Failed to create tempdir")
+            .unwrap();
+        let cache_dir = tempdir.path().join("cache");
+        fs::create_dir(&cache_dir).unwrap();
+
+        config.fallback_cache.dir = cache_dir;
+
+        // Test Read Write
+        config.fallback_cache.rw_mode = CacheModeConfig::ReadWrite;
+
+        {
+            let cache = storage_from_config(&config, runtime.handle()).unwrap();
+
+            runtime.block_on(async move {
+                cache.put("test1", CacheWrite::default()).await.unwrap();
+                cache
+                    .put_preprocessor_cache_entry("test1", PreprocessorCacheEntry::default())
+                    .await
+                    .unwrap();
+            });
+        }
+
+        // Test Read-only
+        config.fallback_cache.rw_mode = CacheModeConfig::ReadOnly;
+
+        {
+            let cache = storage_from_config(&config, runtime.handle()).unwrap();
+
+            runtime.block_on(async move {
+                assert_eq!(
+                    cache
+                        .put("test1", CacheWrite::default())
+                        .await
+                        .unwrap_err()
+                        .to_string(),
+                    "Cannot write to a read-only cache"
+                );
+                assert_eq!(
+                    cache
+                        .put_preprocessor_cache_entry("test1", PreprocessorCacheEntry::default())
+                        .await
+                        .unwrap_err()
+                        .to_string(),
+                    "Cannot write to a read-only cache"
+                );
+            });
+        }
+    }
 }
