@@ -23,6 +23,9 @@ pub mod gcs;
 pub mod gha;
 #[cfg(feature = "memcached")]
 pub mod memcached;
+#[cfg(feature = "oss")]
+pub mod oss;
+pub mod readonly;
 #[cfg(feature = "redis")]
 pub mod redis;
 #[cfg(feature = "s3")]
@@ -30,6 +33,16 @@ pub mod s3;
 #[cfg(feature = "webdav")]
 pub mod webdav;
 
+#[cfg(any(
+    feature = "azure",
+    feature = "gcs",
+    feature = "gha",
+    feature = "s3",
+    feature = "webdav",
+    feature = "oss"
+))]
+pub(crate) mod http_client;
+
 #[cfg(feature = "concurrent-cache")]
 pub mod conccache;
 
