@@ -13,14 +13,13 @@
 // limitations under the License.
 
 use crate::compiler::args::*;
-use crate::compiler::c::{
-    ArtifactDescriptor, CCompilerImpl, CCompilerKind, Language, ParsedArguments,
-};
+use crate::compiler::c::{ArtifactDescriptor, CCompilerImpl, CCompilerKind, ParsedArguments};
 use crate::compiler::{
-    clang, gcc, write_temp_file, Cacheable, ColorMode, CompileCommand, CompilerArguments,
+    clang, gcc, write_temp_file, CCompileCommand, Cacheable, ColorMode, CompileCommand,
+    CompilerArguments, Language, SingleCompileCommand,
 };
 use crate::mock_command::{CommandCreatorSync, RunCommand};
-use crate::util::{run_input_output, OsStrExt};
+use crate::util::{encode_path, run_input_output, OsStrExt};
 use crate::{counted_array, dist};
 use async_trait::async_trait;
 use fs::File;
@@ -60,6 +59,7 @@ impl CCompilerImpl for Msvc {
         &self,
         arguments: &[OsString],
         cwd: &Path,
+        _env_vars: &[(OsString, OsString)],
     ) -> CompilerArguments<ParsedArguments> {
         parse_arguments(arguments, cwd, self.is_clang)
     }
@@ -74,6 +74,7 @@ impl CCompilerImpl for Msvc {
         env_vars: &[(OsString, OsString)],
         may_dist: bool,
         rewrite_includes_only: bool,
+        _preprocessor_cache_mode: bool,
     ) -> Result<process::Output>
     where
         T: CommandCreatorSync,
@@ -92,7 +93,7 @@ impl CCompilerImpl for Msvc {
         .await
     }
 
-    fn generate_compile_commands(
+    fn generate_compile_commands<T>(
         &self,
         path_transformer: &mut dist::PathTransformer,
         executable: &Path,
@@ -100,8 +101,19 @@ impl CCompilerImpl for Msvc {
         cwd: &Path,
         env_vars: &[(OsString, OsString)],
         _rewrite_includes_only: bool,
-    ) -> Result<(CompileCommand, Option<dist::CompileCommand>, Cacheable)> {
-        generate_compile_commands(path_transformer, executable, parsed_args, cwd, env_vars)
+    ) -> Result<(
+        Box<dyn CompileCommand<T>>,
+        Option<dist::CompileCommand>,
+        Cacheable,
+    )>
+    where
+        T: CommandCreatorSync,
+    {
+        generate_compile_commands(path_transformer, executable, parsed_args, cwd, env_vars).map(
+            |(command, dist_command, cacheable)| {
+                (CCompileCommand::new(command), dist_command, cacheable)
+            },
+        )
     }
 }
 
@@ -113,8 +125,10 @@ fn from_local_codepage(multi_byte_str: &[u8]) -> io::Result<String> {
 
 #[cfg(windows)]
 pub fn from_local_codepage(multi_byte_str: &[u8]) -> io::Result<String> {
-    let codepage = winapi::um::winnls::CP_OEMCP;
-    let flags = winapi::um::winnls::MB_ERR_INVALID_CHARS;
+    use windows_sys::Win32::Globalization::{MultiByteToWideChar, CP_OEMCP, MB_ERR_INVALID_CHARS};
+
+    let codepage = CP_OEMCP;
+    let flags = MB_ERR_INVALID_CHARS;
 
     // Empty string
     if multi_byte_str.is_empty() {
@@ -122,7 +136,7 @@ pub fn from_local_codepage(multi_byte_str: &[u8]) -> io::Result<String> {
     }
     unsafe {
         // Get length of UTF-16 string
-        let len = winapi::um::stringapiset::MultiByteToWideChar(
+        let len = MultiByteToWideChar(
             codepage,
             flags,
             multi_byte_str.as_ptr() as _,
@@ -133,7 +147,7 @@ pub fn from_local_codepage(multi_byte_str: &[u8]) -> io::Result<String> {
         if len > 0 {
             // Convert to UTF-16
             let mut wstr: Vec<u16> = Vec::with_capacity(len as usize);
-            let len = winapi::um::stringapiset::MultiByteToWideChar(
+            let len = MultiByteToWideChar(
                 codepage,
                 flags,
                 multi_byte_str.as_ptr() as _,
@@ -238,70 +252,6 @@ where
     bail!("Failed to detect showIncludes prefix")
 }
 
-#[cfg(unix)]
-fn encode_path(dst: &mut dyn Write, path: &Path) -> io::Result<()> {
-    use std::os::unix::prelude::*;
-
-    let bytes = path.as_os_str().as_bytes();
-    dst.write_all(bytes)
-}
-
-#[cfg(windows)]
-fn encode_path(dst: &mut dyn Write, path: &Path) -> io::Result<()> {
-    use std::os::windows::prelude::*;
-
-    let points = path.as_os_str().encode_wide().collect::<Vec<_>>();
-    let bytes = wide_char_to_multi_byte(&points)?; // use_default_char_flag
-    dst.write_all(&bytes)
-}
-
-#[cfg(windows)]
-pub fn wide_char_to_multi_byte(wide_char_str: &[u16]) -> io::Result<Vec<u8>> {
-    let codepage = winapi::um::winnls::CP_OEMCP;
-    let flags = 0;
-    // Empty string
-    if wide_char_str.is_empty() {
-        return Ok(Vec::new());
-    }
-    unsafe {
-        // Get length of multibyte string
-        let len = winapi::um::stringapiset::WideCharToMultiByte(
-            codepage,
-            flags,
-            wide_char_str.as_ptr(),
-            wide_char_str.len() as i32,
-            std::ptr::null_mut(),
-            0,
-            std::ptr::null(),
-            std::ptr::null_mut(),
-        );
-
-        if len > 0 {
-            // Convert from UTF-16 to multibyte
-            let mut astr: Vec<u8> = Vec::with_capacity(len as usize);
-            let len = winapi::um::stringapiset::WideCharToMultiByte(
-                codepage,
-                flags,
-                wide_char_str.as_ptr(),
-                wide_char_str.len() as i32,
-                astr.as_mut_ptr() as _,
-                len,
-                std::ptr::null(),
-                std::ptr::null_mut(),
-            );
-            if len > 0 {
-                astr.set_len(len as usize);
-                if (len as usize) == astr.len() {
-                    return Ok(astr);
-                } else {
-                    return Ok(astr[0..(len as usize)].to_vec());
-                }
-            }
-        }
-        Err(io::Error::last_os_error())
-    }
-}
-
 ArgData! {
     TooHardFlag,
     TooHard(OsString),
@@ -427,6 +377,7 @@ msvc_args!(static ARGS: [ArgInfo<ArgData>; _] = [
     msvc_flag!("Qfast_transcendentals", PassThrough),
     msvc_flag!("Qimprecise_fwaits", PassThrough),
     msvc_flag!("Qpar", PassThrough),
+    msvc_flag!("Qpar-", PassThrough),
     msvc_flag!("Qsafe_fp_loads", PassThrough),
     msvc_flag!("Qspectre", PassThrough),
     msvc_flag!("Qspectre-load", PassThrough),
@@ -557,6 +508,7 @@ pub fn parse_arguments(
 ) -> CompilerArguments<ParsedArguments> {
     let mut output_arg = None;
     let mut input_arg = None;
+    let mut double_dash_input = false;
     let mut common_args = vec![];
     let mut unhashed_args = vec![];
     let mut preprocessor_args = vec![];
@@ -577,7 +529,10 @@ pub fn parse_arguments(
     // Custom iterator to expand `@` arguments which stand for reading a file
     // and interpreting it as a list of more arguments.
     let it = ExpandIncludeFile::new(cwd, arguments);
-    let it = ArgsIter::new(it, (&ARGS[..], &SLASH_ARGS[..]));
+    let mut it = ArgsIter::new(it, (&ARGS[..], &SLASH_ARGS[..]));
+    if is_clang {
+        it = it.with_double_dashes();
+    }
     for arg in it {
         let arg = try_or_cannot_cache!(arg, "argument parse");
         match arg.get_data() {
@@ -618,6 +573,11 @@ pub fn parse_arguments(
             Some(Clang(s)) => clangs.push(s.clone()),
             None => {
                 match arg {
+                    Argument::Raw(ref val) if val == "--" => {
+                        if input_arg.is_none() {
+                            double_dash_input = true;
+                        }
+                    }
                     Argument::Raw(ref val) => {
                         if input_arg.is_some() {
                             // Can't cache compilations with multiple inputs.
@@ -719,8 +679,9 @@ pub fn parse_arguments(
                 | Some(PassThrough(_))
                 | Some(PassThroughPath(_))
                 | Some(PedanticFlag)
-                | Some(Standard(_)) => &mut common_args,
-                Some(Unhashed(_)) => &mut unhashed_args,
+                | Some(Standard(_))
+                | Some(SerializeDiagnostics(_)) => &mut common_args,
+                Some(UnhashedFlag) | Some(Unhashed(_)) => &mut unhashed_args,
 
                 Some(ProfileGenerate) => {
                     profile_generate = true;
@@ -807,6 +768,39 @@ pub fn parse_arguments(
             }
         }
     }
+    if language == Language::Cxx {
+        if let Some(obj) = outputs.get("obj") {
+            // MSVC can produce "type library headers"[1], with the extensions "tlh" and "tli".
+            // These files can be used in later compilation steps to interact with COM interfaces.
+            //
+            // These files are only created when the `#import` directive is used.
+            // Figuring out if an import directive is used would require parsing C++, which would be a lot of work.
+            // To avoid that problem, we just optionally cache these headers if they happen to be produced.
+            // This isn't perfect, but it is easy!
+            //
+            // [1]: https://learn.microsoft.com/en-us/cpp/preprocessor/hash-import-directive-cpp?view=msvc-170#_predir_the_23import_directive_header_files_created_by_import
+            let tlh = obj.path.with_extension("tlh");
+            let tli = obj.path.with_extension("tli");
+
+            // Primary type library header
+            outputs.insert(
+                "tlh",
+                ArtifactDescriptor {
+                    path: tlh,
+                    optional: true,
+                },
+            );
+
+            // Secondary type library header
+            outputs.insert(
+                "tli",
+                ArtifactDescriptor {
+                    path: tli,
+                    optional: true,
+                },
+            );
+        }
+    }
     // -Fd is not taken into account unless -Zi or -ZI are given
     // Clang is currently unable to generate PDB files
     if debug_info && !is_clang {
@@ -829,6 +823,7 @@ pub fn parse_arguments(
 
     CompilerArguments::Ok(ParsedArguments {
         input: input.into(),
+        double_dash_input,
         language,
         compilation_flag,
         depfile,
@@ -838,12 +833,14 @@ pub fn parse_arguments(
         common_args,
         arch_args: vec![],
         unhashed_args,
+        extra_dist_files: vec![],
         extra_hash_files,
         msvc_show_includes: show_includes,
         profile_generate,
         // FIXME: implement color_mode for msvc.
         color_mode: ColorMode::Auto,
         suppress_rewrite_includes_only: false,
+        too_hard_for_preprocessor_cache_mode: None,
     })
 }
 
@@ -852,10 +849,10 @@ fn normpath(path: &str) -> String {
     use std::os::windows::ffi::OsStringExt;
     use std::os::windows::io::AsRawHandle;
     use std::ptr;
-    use winapi::um::fileapi::GetFinalPathNameByHandleW;
+    use windows_sys::Win32::Storage::FileSystem::GetFinalPathNameByHandleW;
     File::open(path)
         .and_then(|f| {
-            let handle = f.as_raw_handle();
+            let handle = f.as_raw_handle() as _;
             let size = unsafe { GetFinalPathNameByHandleW(handle, ptr::null_mut(), 0, 0) };
             if size == 0 {
                 return Err(io::Error::last_os_error());
@@ -882,6 +879,64 @@ fn normpath(path: &str) -> String {
     path.to_owned()
 }
 
+#[allow(clippy::too_many_arguments)]
+pub fn preprocess_cmd<T>(
+    cmd: &mut T,
+    parsed_args: &ParsedArguments,
+    cwd: &Path,
+    env_vars: &[(OsString, OsString)],
+    may_dist: bool,
+    rewrite_includes_only: bool,
+    is_clang: bool,
+) where
+    T: RunCommand,
+{
+    // When performing distributed compilation, line number info is important for error
+    // reporting and to not cause spurious compilation failure (e.g. no exceptions build
+    // fails due to exceptions transitively included in the stdlib).
+    // With -fprofile-generate line number information is important, so use -E.
+    // Otherwise, use -EP to maximize cache hits (because no absolute file paths are
+    // emitted) and improve performance.
+    if may_dist || parsed_args.profile_generate {
+        cmd.arg("-E");
+    } else {
+        cmd.arg("-EP");
+    }
+
+    cmd.arg("-nologo")
+        .args(&parsed_args.preprocessor_args)
+        .args(&parsed_args.dependency_args)
+        .args(&parsed_args.common_args)
+        .env_clear()
+        .envs(env_vars.to_vec())
+        .current_dir(cwd);
+
+    if is_clang {
+        if parsed_args.depfile.is_some() && !parsed_args.msvc_show_includes {
+            cmd.arg("-showIncludes");
+        }
+    } else {
+        // cl.exe can product the dep list itself, in a JSON format that some tools will be expecting.
+        if let Some(ref depfile) = parsed_args.depfile {
+            cmd.arg("/sourceDependencies");
+            cmd.arg(depfile);
+        }
+        // Windows SDK generates C4668 during preprocessing, but compiles fine.
+        // Read for more info: https://github.com/mozilla/sccache/issues/1725
+        // And here: https://github.com/mozilla/sccache/issues/2250
+        cmd.arg("/WX-");
+    }
+
+    if rewrite_includes_only && is_clang {
+        cmd.arg("-clang:-frewrite-includes");
+    }
+
+    if parsed_args.double_dash_input {
+        cmd.arg("--");
+    }
+    cmd.arg(&parsed_args.input);
+}
+
 #[allow(clippy::too_many_arguments)]
 pub async fn preprocess<T>(
     creator: &T,
@@ -898,33 +953,15 @@ where
     T: CommandCreatorSync,
 {
     let mut cmd = creator.clone().new_command_sync(executable);
-
-    // When performing distributed compilation, line number info is important for error
-    // reporting and to not cause spurious compilation failure (e.g. no exceptions build
-    // fails due to exceptions transitively included in the stdlib).
-    // With -fprofile-generate line number information is important, so use -E.
-    // Otherwise, use -EP to maximize cache hits (because no absolute file paths are
-    // emitted) and improve performance.
-    if may_dist || parsed_args.profile_generate {
-        cmd.arg("-E");
-    } else {
-        cmd.arg("-EP");
-    }
-
-    cmd.arg(&parsed_args.input)
-        .arg("-nologo")
-        .args(&parsed_args.preprocessor_args)
-        .args(&parsed_args.dependency_args)
-        .args(&parsed_args.common_args)
-        .env_clear()
-        .envs(env_vars.iter().map(|&(ref k, ref v)| (k, v)))
-        .current_dir(cwd);
-    if parsed_args.depfile.is_some() && !parsed_args.msvc_show_includes {
-        cmd.arg("-showIncludes");
-    }
-    if rewrite_includes_only && is_clang {
-        cmd.arg("-clang:-frewrite-includes");
-    }
+    preprocess_cmd(
+        &mut cmd,
+        parsed_args,
+        cwd,
+        env_vars,
+        may_dist,
+        rewrite_includes_only,
+        is_clang,
+    );
 
     if log_enabled!(Debug) {
         debug!("preprocess: {:?}", cmd);
@@ -936,9 +973,12 @@ where
 
     let output = run_input_output(cmd, None).await?;
 
+    if !is_clang {
+        return Ok(output);
+    }
+
     let parsed_args = &parsed_args;
-    if let (Some(obj), &Some(ref depfile)) = (parsed_args.outputs.get("obj"), &parsed_args.depfile)
-    {
+    if let (Some(obj), Some(depfile)) = (parsed_args.outputs.get("obj"), &parsed_args.depfile) {
         let objfile = &obj.path;
         let f = File::create(cwd.join(depfile))?;
         let mut f = BufWriter::new(f);
@@ -1001,7 +1041,11 @@ fn generate_compile_commands(
     parsed_args: &ParsedArguments,
     cwd: &Path,
     env_vars: &[(OsString, OsString)],
-) -> Result<(CompileCommand, Option<dist::CompileCommand>, Cacheable)> {
+) -> Result<(
+    SingleCompileCommand,
+    Option<dist::CompileCommand>,
+    Cacheable,
+)> {
     #[cfg(not(feature = "dist-client"))]
     let _ = path_transformer;
 
@@ -1028,17 +1072,16 @@ fn generate_compile_commands(
     let mut fo = OsString::from("-Fo");
     fo.push(out_file);
 
-    let mut arguments: Vec<OsString> = vec![
-        parsed_args.compilation_flag.clone(),
-        parsed_args.input.clone().into(),
-        fo,
-    ];
-    arguments.extend(parsed_args.preprocessor_args.clone());
-    arguments.extend(parsed_args.dependency_args.clone());
-    arguments.extend(parsed_args.unhashed_args.clone());
-    arguments.extend(parsed_args.common_args.clone());
-
-    let command = CompileCommand {
+    let mut arguments: Vec<OsString> = vec![parsed_args.compilation_flag.clone(), fo];
+    arguments.extend_from_slice(&parsed_args.preprocessor_args);
+    arguments.extend_from_slice(&parsed_args.dependency_args);
+    arguments.extend_from_slice(&parsed_args.unhashed_args);
+    arguments.extend_from_slice(&parsed_args.common_args);
+    if parsed_args.double_dash_input {
+        arguments.push("--".into());
+    }
+    arguments.push(parsed_args.input.clone().into());
+    let command = SingleCompileCommand {
         executable: executable.to_owned(),
         arguments,
         env_vars: env_vars.to_owned(),
@@ -1054,17 +1097,19 @@ fn generate_compile_commands(
         let mut fo = String::from("-Fo");
         fo.push_str(&path_transformer.as_dist(out_file)?);
 
-        let mut arguments: Vec<String> = vec![
-            parsed_args.compilation_flag.clone().into_string().ok()?,
-            path_transformer.as_dist(&parsed_args.input)?,
-            fo,
-        ];
+        let mut arguments: Vec<String> =
+            vec![parsed_args.compilation_flag.clone().into_string().ok()?, fo];
         // It's important to avoid preprocessor_args because of things like /FI which
         // forcibly includes another file. This does mean we're potentially vulnerable
         // to misidentification of flags like -DYNAMICBASE (though in that specific
         // case we're safe as it only applies to link time, which sccache avoids).
         arguments.extend(dist::osstrings_to_strings(&parsed_args.common_args)?);
 
+        if parsed_args.double_dash_input {
+            arguments.push("--".into());
+        }
+        arguments.push(path_transformer.as_dist(&parsed_args.input)?);
+
         Some(dist::CompileCommand {
             executable: path_transformer.as_dist(executable)?,
             arguments,
@@ -1139,23 +1184,20 @@ impl<'a> ExpandIncludeFile<'a> {
     }
 }
 
-impl<'a> Iterator for ExpandIncludeFile<'a> {
+impl Iterator for ExpandIncludeFile<'_> {
     type Item = OsString;
 
     fn next(&mut self) -> Option<OsString> {
         loop {
             // Visit all arguments found in the most recently read response file.
             // Since response files are not recursive, we do not need to worry
-            // about these containing addditional @ directives.
+            // about these containing additional @ directives.
             if let Some(response_file_arg) = self.stack.pop() {
                 return Some(response_file_arg);
             }
 
             // Visit the next argument provided by the original command iterator.
-            let arg = match self.args.pop() {
-                Some(arg) => arg,
-                None => return None,
-            };
+            let arg = self.args.pop()?;
             let file_arg = match arg.split_prefix("@") {
                 Some(file_arg) => file_arg,
                 None => return Some(arg),
@@ -1166,7 +1208,7 @@ impl<'a> Iterator for ExpandIncludeFile<'a> {
                 Ok(content) => content,
                 Err(err) => {
                     debug!("failed to read @-file `{}`: {}", file_path.display(), err);
-                    // If we failed to read the file content, return the orginal arg (including the `@` directive).
+                    // If we failed to read the file content, return the original arg (including the `@` directive).
                     return Some(arg);
                 }
             };
@@ -1191,13 +1233,16 @@ where
     let mut buf = Vec::new();
     reader.read_to_end(&mut buf)?;
 
-    let (result, _) = encoding::decode(
-        &buf,
-        encoding::DecoderTrap::Strict,
-        encoding::all::ISO_8859_1,
-    );
+    let (result, _, has_error) = encoding_rs::WINDOWS_1252.decode(&buf);
 
-    result.map_err(|err| io::Error::new(io::ErrorKind::Other, err.into_owned()))
+    if has_error {
+        Err(io::Error::new(
+            io::ErrorKind::InvalidData,
+            "failed to decode text",
+        ))
+    } else {
+        Ok(result.to_string())
+    }
 }
 
 /// An iterator over the arguments in a Windows command line.
@@ -1242,7 +1287,7 @@ where
     }
 }
 
-impl<'a> SplitMsvcResponseFileArgs<'a> {
+impl SplitMsvcResponseFileArgs<'_> {
     /// Appends backslashes to `target` by decrementing `count`.
     /// If `step` is >1, then `count` is decremented by `step`, resulting in 1 backslash appended for every `step`.
     fn append_backslashes_to(target: &mut String, count: &mut usize, step: usize) {
@@ -1253,7 +1298,7 @@ impl<'a> SplitMsvcResponseFileArgs<'a> {
     }
 }
 
-impl<'a> Iterator for SplitMsvcResponseFileArgs<'a> {
+impl Iterator for SplitMsvcResponseFileArgs<'_> {
     type Item = String;
 
     fn next(&mut self) -> Option<String> {
@@ -1330,12 +1375,18 @@ mod test {
     use super::*;
     use crate::compiler::*;
     use crate::mock_command::*;
+    use crate::server;
+    use crate::test::mock_storage::MockStorage;
     use crate::test::utils::*;
 
     fn parse_arguments(arguments: Vec<OsString>) -> CompilerArguments<ParsedArguments> {
         super::parse_arguments(&arguments, &std::env::current_dir().unwrap(), false)
     }
 
+    fn parse_arguments_clang(arguments: Vec<OsString>) -> CompilerArguments<ParsedArguments> {
+        super::parse_arguments(&arguments, &std::env::current_dir().unwrap(), true)
+    }
+
     #[test]
     fn test_detect_showincludes_prefix() {
         drop(env_logger::try_init());
@@ -1393,6 +1444,72 @@ mod test {
         assert!(!msvc_show_includes);
     }
 
+    #[test]
+    fn test_cpp_parse_arguments_collects_type_library_headers() {
+        let args = ovec!["-c", "foo.cpp", "-Fofoo.obj"];
+        let ParsedArguments {
+            input,
+            language,
+            outputs,
+            ..
+        } = match parse_arguments(args) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Got unexpected parse result: {:?}", o),
+        };
+        assert_eq!(Some("foo.cpp"), input.to_str());
+        assert_eq!(Language::Cxx, language);
+        assert_map_contains!(
+            outputs,
+            (
+                "obj",
+                ArtifactDescriptor {
+                    path: PathBuf::from("foo.obj"),
+                    optional: false
+                }
+            ),
+            (
+                "tlh",
+                ArtifactDescriptor {
+                    path: PathBuf::from("foo.tlh"),
+                    optional: true
+                }
+            ),
+            (
+                "tli",
+                ArtifactDescriptor {
+                    path: PathBuf::from("foo.tli"),
+                    optional: true
+                }
+            )
+        );
+    }
+
+    #[test]
+    fn test_c_parse_arguments_does_not_collect_type_library_headers() {
+        let args = ovec!["-c", "foo.c", "-Fofoo.obj"];
+        let ParsedArguments {
+            input,
+            language,
+            outputs,
+            ..
+        } = match parse_arguments(args) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Got unexpected parse result: {:?}", o),
+        };
+        assert_eq!(Some("foo.c"), input.to_str());
+        assert_eq!(Language::C, language);
+        assert_map_contains!(
+            outputs,
+            (
+                "obj",
+                ArtifactDescriptor {
+                    path: PathBuf::from("foo.obj"),
+                    optional: false
+                }
+            )
+        );
+    }
+
     #[test]
     fn test_parse_compile_flag() {
         let args = ovec!["/c", "foo.c", "-Fofoo.obj"];
@@ -1459,6 +1576,65 @@ mod test {
         assert!(!msvc_show_includes);
     }
 
+    #[test]
+    fn test_parse_arguments_double_dash() {
+        let args = ovec!["-c", "-Fofoo.obj", "--", "foo.c"];
+        let ParsedArguments {
+            input,
+            double_dash_input,
+            common_args,
+            ..
+        } = match parse_arguments(args.clone()) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Got unexpected parse result: {:?}", o),
+        };
+        assert_eq!(Some("foo.c"), input.to_str());
+        // MSVC doesn't support double dashes. If we got one, we'll pass them
+        // through to MSVC for it to error out.
+        assert!(!double_dash_input);
+        assert_eq!(ovec!["--"], common_args);
+
+        let ParsedArguments {
+            input,
+            double_dash_input,
+            common_args,
+            ..
+        } = match parse_arguments_clang(args) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Got unexpected parse result: {:?}", o),
+        };
+        assert_eq!(Some("foo.c"), input.to_str());
+        assert!(double_dash_input);
+        assert!(common_args.is_empty());
+
+        let args = ovec!["-c", "-Fofoo.obj", "foo.c", "--"];
+        let ParsedArguments {
+            input,
+            double_dash_input,
+            common_args,
+            ..
+        } = match parse_arguments_clang(args) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Got unexpected parse result: {:?}", o),
+        };
+        assert_eq!(Some("foo.c"), input.to_str());
+        // Double dash after input file is ignored.
+        assert!(!double_dash_input);
+        assert!(common_args.is_empty());
+
+        let args = ovec!["-c", "-Fofoo.obj", "foo.c", "--", "bar.c"];
+        assert_eq!(
+            CompilerArguments::CannotCache("multiple input files", Some("[\"bar.c\"]".to_string())),
+            parse_arguments_clang(args)
+        );
+
+        let args = ovec!["-c", "-Fofoo.obj", "foo.c", "--", "-fPIC"];
+        assert_eq!(
+            CompilerArguments::CannotCache("multiple input files", Some("[\"-fPIC\"]".to_string())),
+            parse_arguments_clang(args)
+        );
+    }
+
     #[test]
     fn parse_argument_slashes() {
         let args = ovec!["-c", "foo.c", "/Fofoo.obj"];
@@ -1494,14 +1670,19 @@ mod test {
     #[test]
     fn parse_deps_arguments() {
         let arg_sets = vec![
-            ovec!["-c", "foo.c", "/Fofoo.obj", "/depsfoo.obj.pp"],
-            ovec!["-c", "foo.c", "/Fofoo.obj", "/sourceDependenciesfoo.obj.pp"],
+            ovec!["-c", "foo.c", "/Fofoo.obj", "/depsfoo.obj.json"],
+            ovec![
+                "-c",
+                "foo.c",
+                "/Fofoo.obj",
+                "/sourceDependenciesfoo.obj.json"
+            ],
             ovec![
                 "-c",
                 "foo.c",
                 "/Fofoo.obj",
                 "/sourceDependencies",
-                "foo.obj.pp"
+                "foo.obj.json"
             ],
         ];
 
@@ -1521,7 +1702,7 @@ mod test {
             };
             assert_eq!(Some("foo.c"), input.to_str());
             assert_eq!(Language::C, language);
-            assert_eq!(Some(PathBuf::from_str("foo.obj.pp").unwrap()), depfile);
+            assert_eq!(Some(PathBuf::from_str("foo.obj.json").unwrap()), depfile);
             assert_map_contains!(
                 outputs,
                 (
@@ -1827,6 +2008,8 @@ mod test {
     fn test_parse_arguments_passthrough() {
         let args = ovec![
             "-Oy",
+            "-Qpar",
+            "-Qpar-",
             "-Gw",
             "-EHa",
             "-Fmdictionary-map",
@@ -1850,7 +2033,7 @@ mod test {
         assert!(!common_args.is_empty());
         assert_eq!(
             common_args,
-            ovec!("-Oy", "-Gw", "-EHa", "-Fmdictionary-map")
+            ovec!("-Oy", "-Qpar", "-Qpar-", "-Gw", "-EHa", "-Fmdictionary-map")
         );
     }
 
@@ -2133,10 +2316,63 @@ mod test {
             .unwrap();
         let cmd_file_path = td.path().join("foo");
         {
-            use encoding::{all::UTF_16LE, EncoderTrap::Strict, Encoding};
             let mut file = File::create(&cmd_file_path).unwrap();
-            let content = UTF_16LE.encode("-c foo€.c -o foo.o", Strict).unwrap();
-            file.write_all(&[0xFF, 0xFE]).unwrap(); // little endian BOM
+            // pre-encoded with utf16le
+            let content: [u8; 0x26] = [
+                0xFF, 0xFE, // little endian BOM
+                // `-c foo€.c -o foo.o`
+                0x2D, 0x00, 0x63, 0x00, 0x20, 0x00, 0x66, 0x00, 0x6F, 0x00, 0x6F, 0x00, 0xAC, 0x20,
+                0x2E, 0x00, 0x63, 0x00, 0x20, 0x00, 0x2D, 0x00, 0x6F, 0x00, 0x20, 0x00, 0x66, 0x00,
+                0x6F, 0x00, 0x6F, 0x00, 0x2E, 0x00, 0x6F, 0x00,
+            ];
+            file.write_all(&content).unwrap();
+        }
+        let arg = format!("@{}", cmd_file_path.display());
+        let ParsedArguments {
+            input,
+            language,
+            outputs,
+            preprocessor_args,
+            msvc_show_includes,
+            common_args,
+            ..
+        } = match parse_arguments(ovec![arg]) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Failed to parse @-file, err: {:?}", o),
+        };
+        assert_eq!(Some("foo€.c"), input.to_str());
+        assert_eq!(Language::C, language);
+        assert_map_contains!(
+            outputs,
+            (
+                "obj",
+                ArtifactDescriptor {
+                    path: "foo.o".into(),
+                    optional: false
+                }
+            )
+        );
+        assert!(preprocessor_args.is_empty());
+        assert!(common_args.is_empty());
+        assert!(!msvc_show_includes);
+    }
+
+    #[test]
+    fn test_responsefile_encoding_win1252() {
+        let td = tempfile::Builder::new()
+            .prefix("sccache")
+            .tempdir()
+            .unwrap();
+        let cmd_file_path = td.path().join("foo");
+        {
+            let mut file = File::create(&cmd_file_path).unwrap();
+            // pre-encoded with Windows 1252
+            let content: [u8; 0x12] = [
+                // `-c foo€.c -o foo.o`
+                // the euro symbol is 0x80 in Windows 1252 (and undefined in ISO-8859-1)
+                0x2D, 0x63, 0x20, 0x66, 0x6F, 0x6F, 0x80, 0x2E, 0x63, 0x20, 0x2D, 0x6F, 0x20, 0x66,
+                0x6F, 0x6F, 0x2E, 0x6F,
+            ];
             file.write_all(&content).unwrap();
         }
         let arg = format!("@{}", cmd_file_path.display());
@@ -2185,12 +2421,29 @@ mod test {
         );
     }
 
+    #[test]
+    fn test_preprocess_double_dash_input() {
+        let args = ovec!["-c", "-Fofoo.o.bj", "--", "foo.c"];
+        let parsed_args = match parse_arguments_clang(args) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Got unexpected parse result: {:?}", o),
+        };
+        let mut cmd = MockCommand {
+            child: None,
+            args: vec![],
+        };
+        preprocess_cmd(&mut cmd, &parsed_args, Path::new(""), &[], true, true, true);
+        let expected_args = ovec!["-E", "-nologo", "-clang:-frewrite-includes", "--", "foo.c"];
+        assert_eq!(cmd.args, expected_args);
+    }
+
     #[test]
     fn test_compile_simple() {
         let creator = new_creator();
         let f = TestFixture::new();
         let parsed_args = ParsedArguments {
             input: "foo.c".into(),
+            double_dash_input: false,
             language: Language::C,
             compilation_flag: "-c".into(),
             depfile: None,
@@ -2208,16 +2461,22 @@ mod test {
             common_args: vec![],
             arch_args: vec![],
             unhashed_args: vec![],
+            extra_dist_files: vec![],
             extra_hash_files: vec![],
             msvc_show_includes: false,
             profile_generate: false,
             color_mode: ColorMode::Auto,
             suppress_rewrite_includes_only: false,
+            too_hard_for_preprocessor_cache_mode: None,
         };
+        let runtime = single_threaded_runtime();
+        let storage = MockStorage::new(None, false);
+        let storage: std::sync::Arc<MockStorage> = std::sync::Arc::new(storage);
+        let service = server::SccacheService::mock_with_storage(storage, runtime.handle().clone());
         let compiler = &f.bins[0];
         // Compiler invocation.
         next_command(&creator, Ok(MockChild::new(exit_status(0), "", "")));
-        let mut path_transformer = dist::PathTransformer::default();
+        let mut path_transformer = dist::PathTransformer::new();
         let (command, dist_command, cacheable) = generate_compile_commands(
             &mut path_transformer,
             compiler,
@@ -2230,12 +2489,34 @@ mod test {
         assert!(dist_command.is_some());
         #[cfg(not(feature = "dist-client"))]
         assert!(dist_command.is_none());
-        let _ = command.execute(&creator).wait();
+        let _ = command.execute(&service, &creator).wait();
         assert_eq!(Cacheable::Yes, cacheable);
         // Ensure that we ran all processes.
         assert_eq!(0, creator.lock().unwrap().children.len());
     }
 
+    #[test]
+    fn test_compile_double_dash_input() {
+        let args = ovec!["-c", "-Fofoo.obj", "--", "foo.c"];
+        let parsed_args = match parse_arguments_clang(args) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Got unexpected parse result: {:?}", o),
+        };
+        let f = TestFixture::new();
+        let compiler = &f.bins[0];
+        let mut path_transformer = dist::PathTransformer::new();
+        let (command, _, _) = generate_compile_commands(
+            &mut path_transformer,
+            compiler,
+            &parsed_args,
+            f.tempdir.path(),
+            &[],
+        )
+        .unwrap();
+        let expected_args = ovec!["-c", "-Fofoo.obj", "--", "foo.c"];
+        assert_eq!(command.arguments, expected_args);
+    }
+
     #[test]
     fn test_compile_not_cacheable_pdb() {
         let creator = new_creator();
@@ -2243,6 +2524,7 @@ mod test {
         let pdb = f.touch("foo.pdb").unwrap();
         let parsed_args = ParsedArguments {
             input: "foo.c".into(),
+            double_dash_input: false,
             language: Language::C,
             compilation_flag: "/c".into(),
             depfile: None,
@@ -2269,16 +2551,22 @@ mod test {
             common_args: vec![],
             arch_args: vec![],
             unhashed_args: vec![],
+            extra_dist_files: vec![],
             extra_hash_files: vec![],
             msvc_show_includes: false,
             profile_generate: false,
             color_mode: ColorMode::Auto,
             suppress_rewrite_includes_only: false,
+            too_hard_for_preprocessor_cache_mode: None,
         };
+        let runtime = single_threaded_runtime();
+        let storage = MockStorage::new(None, false);
+        let storage: std::sync::Arc<MockStorage> = std::sync::Arc::new(storage);
+        let service = server::SccacheService::mock_with_storage(storage, runtime.handle().clone());
         let compiler = &f.bins[0];
         // Compiler invocation.
         next_command(&creator, Ok(MockChild::new(exit_status(0), "", "")));
-        let mut path_transformer = dist::PathTransformer::default();
+        let mut path_transformer = dist::PathTransformer::new();
         let (command, dist_command, cacheable) = generate_compile_commands(
             &mut path_transformer,
             compiler,
@@ -2291,7 +2579,7 @@ mod test {
         assert!(dist_command.is_some());
         #[cfg(not(feature = "dist-client"))]
         assert!(dist_command.is_none());
-        let _ = command.execute(&creator).wait();
+        let _ = command.execute(&service, &creator).wait();
         assert_eq!(Cacheable::No, cacheable);
         // Ensure that we ran all processes.
         assert_eq!(0, creator.lock().unwrap().children.len());
@@ -2324,7 +2612,8 @@ mod test {
     #[test]
     #[cfg(windows)]
     fn local_oem_codepage_conversions() {
-        use winapi::um::winnls::GetOEMCP;
+        use crate::util::wide_char_to_multi_byte;
+        use windows_sys::Win32::Globalization::GetOEMCP;
 
         let current_oemcp = unsafe { GetOEMCP() };
         // We don't control the local OEM codepage so test only if it is one of:
