@@ -12,11 +12,11 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
+use crate::cache::CacheMode;
 use directories::ProjectDirs;
 use fs::File;
 use fs_err as fs;
-use lazy_static::lazy_static;
-use regex::Regex;
+use once_cell::sync::Lazy;
 #[cfg(any(feature = "dist-client", feature = "dist-server"))]
 use serde::ser::Serializer;
 use serde::{
@@ -33,12 +33,11 @@ use std::result::Result as StdResult;
 use std::str::FromStr;
 use std::sync::Mutex;
 
+pub use crate::cache::PreprocessorCacheModeConfig;
 use crate::errors::*;
 
-lazy_static! {
-    static ref CACHED_CONFIG_PATH: PathBuf = CachedConfig::file_config_path();
-    static ref CACHED_CONFIG: Mutex<Option<CachedFileConfig>> = Mutex::new(None);
-}
+static CACHED_CONFIG_PATH: Lazy<PathBuf> = Lazy::new(CachedConfig::file_config_path);
+static CACHED_CONFIG: Mutex<Option<CachedFileConfig>> = Mutex::new(None);
 
 const ORGANIZATION: &str = "Mozilla";
 const APP_NAME: &str = "sccache";
@@ -79,20 +78,19 @@ fn default_toolchain_cache_size() -> u64 {
 }
 
 pub fn parse_size(val: &str) -> Option<u64> {
-    let re = Regex::new(r"^(\d+)([KMGT])$").expect("Fixed regex parse failure");
-    re.captures(val)
-        .and_then(|caps| {
-            caps.get(1)
-                .and_then(|size| u64::from_str(size.as_str()).ok())
-                .map(|size| (size, caps.get(2)))
-        })
-        .and_then(|(size, suffix)| match suffix.map(|s| s.as_str()) {
-            Some("K") => Some(1024 * size),
-            Some("M") => Some(1024 * 1024 * size),
-            Some("G") => Some(1024 * 1024 * 1024 * size),
-            Some("T") => Some(1024 * 1024 * 1024 * 1024 * size),
-            _ => None,
-        })
+    let multiplier = match val.chars().last() {
+        Some('K') => 1024,
+        Some('M') => 1024 * 1024,
+        Some('G') => 1024 * 1024 * 1024,
+        Some('T') => 1024 * 1024 * 1024 * 1024,
+        _ => 1,
+    };
+    let val = if multiplier > 1 && !val.is_empty() {
+        val.split_at(val.len() - 1).0
+    } else {
+        val
+    };
+    u64::from_str(val).ok().map(|size| size * multiplier)
 }
 
 #[cfg(any(feature = "dist-client", feature = "dist-server"))]
@@ -162,6 +160,8 @@ pub struct DiskCacheConfig {
     pub dir: PathBuf,
     // TODO: use deserialize_with to allow human-readable sizes in toml
     pub size: u64,
+    pub preprocessor_cache_mode: PreprocessorCacheModeConfig,
+    pub rw_mode: CacheModeConfig,
 }
 
 impl Default for DiskCacheConfig {
@@ -169,19 +169,30 @@ impl Default for DiskCacheConfig {
         DiskCacheConfig {
             dir: default_disk_cache_dir(),
             size: default_disk_cache_size(),
+            preprocessor_cache_mode: PreprocessorCacheModeConfig::activated(),
+            rw_mode: CacheModeConfig::ReadWrite,
         }
     }
 }
 
 #[derive(Copy, Clone, Debug, PartialEq, Eq, Serialize, Deserialize)]
 #[serde(deny_unknown_fields)]
-pub enum GCSCacheRWMode {
+pub enum CacheModeConfig {
     #[serde(rename = "READ_ONLY")]
     ReadOnly,
     #[serde(rename = "READ_WRITE")]
     ReadWrite,
 }
 
+impl From<CacheModeConfig> for CacheMode {
+    fn from(value: CacheModeConfig) -> Self {
+        match value {
+            CacheModeConfig::ReadOnly => CacheMode::ReadOnly,
+            CacheModeConfig::ReadWrite => CacheMode::ReadWrite,
+        }
+    }
+}
+
 #[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
 #[serde(deny_unknown_fields)]
 pub struct GCSCacheConfig {
@@ -189,7 +200,7 @@ pub struct GCSCacheConfig {
     pub key_prefix: String,
     pub cred_path: Option<String>,
     pub service_account: Option<String>,
-    pub rw_mode: GCSCacheRWMode,
+    pub rw_mode: CacheModeConfig,
     pub credential_url: Option<String>,
 }
 
@@ -211,27 +222,82 @@ pub struct GHACacheConfig {
 /// Please change this value freely if we have a better choice.
 const DEFAULT_MEMCACHED_CACHE_EXPIRATION: u32 = 86400;
 
-#[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
+fn default_memcached_cache_expiration() -> u32 {
+    DEFAULT_MEMCACHED_CACHE_EXPIRATION
+}
+
+#[derive(Debug, PartialEq, Eq, Serialize, Deserialize, Default)]
 #[serde(deny_unknown_fields)]
 pub struct MemcachedCacheConfig {
+    #[serde(alias = "endpoint")]
     pub url: String,
+
+    /// Username to authenticate with.
+    pub username: Option<String>,
+
+    /// Password to authenticate with.
+    pub password: Option<String>,
+
     /// the expiration time in seconds.
     ///
     /// Default to 24 hours (86400)
     /// Up to 30 days (2592000)
+    #[serde(default = "default_memcached_cache_expiration")]
     pub expiration: u32,
+
+    #[serde(default)]
+    pub key_prefix: String,
 }
 
-#[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
+/// redis has no default TTL - all caches live forever
+///
+/// We keep the TTL as 0 here as redis does
+///
+/// Please change this value freely if we have a better choice.
+const DEFAULT_REDIS_CACHE_TTL: u64 = 0;
+pub const DEFAULT_REDIS_DB: u32 = 0;
+#[derive(Debug, PartialEq, Eq, Serialize, Deserialize, Default)]
 #[serde(deny_unknown_fields)]
 pub struct RedisCacheConfig {
-    pub url: String,
+    /// The single-node redis endpoint.
+    /// Mutually exclusive with `cluster_endpoints`.
+    pub endpoint: Option<String>,
+
+    /// The redis cluster endpoints.
+    /// Mutually exclusive with `endpoint`.
+    pub cluster_endpoints: Option<String>,
+
+    /// Username to authenticate with.
+    pub username: Option<String>,
+
+    /// Password to authenticate with.
+    pub password: Option<String>,
+
+    /// The redis URL.
+    /// Deprecated in favor of `endpoint`.
+    pub url: Option<String>,
+
+    /// the db number to use
+    ///
+    /// Default to 0
+    #[serde(default)]
+    pub db: u32,
+
+    /// the ttl (expiration) time in seconds.
+    ///
+    /// Default to infinity (0)
+    #[serde(default, alias = "expiration")]
+    pub ttl: u64,
+
+    #[serde(default)]
+    pub key_prefix: String,
 }
 
 #[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
 #[serde(deny_unknown_fields)]
 pub struct WebdavCacheConfig {
     pub endpoint: String,
+    #[serde(default)]
     pub key_prefix: String,
     pub username: Option<String>,
     pub password: Option<String>,
@@ -243,10 +309,23 @@ pub struct WebdavCacheConfig {
 pub struct S3CacheConfig {
     pub bucket: String,
     pub region: Option<String>,
+    #[serde(default)]
     pub key_prefix: String,
     pub no_credentials: bool,
     pub endpoint: Option<String>,
     pub use_ssl: Option<bool>,
+    pub server_side_encryption: Option<bool>,
+    pub enable_virtual_host_style: Option<bool>,
+}
+
+#[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
+#[serde(deny_unknown_fields)]
+pub struct OSSCacheConfig {
+    pub bucket: String,
+    #[serde(default)]
+    pub key_prefix: String,
+    pub endpoint: Option<String>,
+    pub no_credentials: bool,
 }
 
 #[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
@@ -267,6 +346,7 @@ pub enum CacheType {
     Redis(RedisCacheConfig),
     S3(S3CacheConfig),
     Webdav(WebdavCacheConfig),
+    OSS(OSSCacheConfig),
     ConcurrentDisk(ConcurrentDiskCacheConfig),
 }
 
@@ -281,6 +361,7 @@ pub struct CacheConfigs {
     pub redis: Option<RedisCacheConfig>,
     pub s3: Option<S3CacheConfig>,
     pub webdav: Option<WebdavCacheConfig>,
+    pub oss: Option<OSSCacheConfig>,
     pub concdisk: Option<ConcurrentDiskCacheConfig>,
 }
 
@@ -297,6 +378,7 @@ impl CacheConfigs {
             redis,
             s3,
             webdav,
+            oss,
             concdisk,
         } = self;
 
@@ -308,7 +390,8 @@ impl CacheConfigs {
             .or_else(|| gcs.map(CacheType::GCS))
             .or_else(|| gha.map(CacheType::GHA))
             .or_else(|| azure.map(CacheType::Azure))
-            .or_else(|| webdav.map(CacheType::Webdav));
+            .or_else(|| webdav.map(CacheType::Webdav))
+            .or_else(|| oss.map(CacheType::OSS));
 
         let fallback = disk.unwrap_or_default();
 
@@ -326,6 +409,7 @@ impl CacheConfigs {
             redis,
             s3,
             webdav,
+            oss,
             concdisk,
         } = other;
 
@@ -353,6 +437,9 @@ impl CacheConfigs {
         if webdav.is_some() {
             self.webdav = webdav
         }
+        if oss.is_some() {
+            self.oss = oss
+        }
         if concdisk.is_some() {
             self.concdisk = concdisk
         }
@@ -509,7 +596,7 @@ pub fn try_read_config_file<T: DeserializeOwned>(path: &Path) -> Result<Option<T
         }
     }
 
-    let res = if path.extension().map_or(false, |e| e == "json") {
+    let res = if path.extension().is_some_and(|e| e == "json") {
         serde_json::from_str(&string)
             .with_context(|| format!("Failed to load json config file from {}", path.display()))?
     } else {
@@ -525,27 +612,52 @@ pub struct EnvConfig {
     cache: CacheConfigs,
 }
 
+fn key_prefix_from_env_var(env_var_name: &str) -> String {
+    env::var(env_var_name)
+        .ok()
+        .as_ref()
+        .map(|s| s.trim_end_matches('/'))
+        .filter(|s| !s.is_empty())
+        .unwrap_or_default()
+        .to_owned()
+}
+
+fn number_from_env_var<A: std::str::FromStr>(env_var_name: &str) -> Option<Result<A>>
+where
+    <A as FromStr>::Err: std::fmt::Debug,
+{
+    let value = env::var(env_var_name).ok()?;
+
+    value
+        .parse::<A>()
+        .map_err(|err| anyhow!("{env_var_name} value is invalid: {err:?}"))
+        .into()
+}
+
+fn bool_from_env_var(env_var_name: &str) -> Result<Option<bool>> {
+    env::var(env_var_name)
+        .ok()
+        .map(|value| match value.to_lowercase().as_str() {
+            "true" | "on" | "1" => Ok(true),
+            "false" | "off" | "0" => Ok(false),
+            _ => bail!(
+                "{} must be 'true', 'on', '1', 'false', 'off' or '0'.",
+                env_var_name
+            ),
+        })
+        .transpose()
+}
+
 fn config_from_env() -> Result<EnvConfig> {
     // ======= AWS =======
     let s3 = if let Ok(bucket) = env::var("SCCACHE_BUCKET") {
         let region = env::var("SCCACHE_REGION").ok();
-        let no_credentials =
-            env::var("SCCACHE_S3_NO_CREDENTIALS").map_or(Ok(false), |val| match val.as_str() {
-                "true" | "1" => Ok(true),
-                "false" | "0" => Ok(false),
-                _ => bail!("SCCACHE_S3_NO_CREDENTIALS must be 'true', '1', 'false', or '0'."),
-            })?;
-        let use_ssl = env::var("SCCACHE_S3_USE_SSL")
-            .ok()
-            .map(|value| value != "off");
+        let no_credentials = bool_from_env_var("SCCACHE_S3_NO_CREDENTIALS")?.unwrap_or(false);
+        let use_ssl = bool_from_env_var("SCCACHE_S3_USE_SSL")?;
+        let server_side_encryption = bool_from_env_var("SCCACHE_S3_SERVER_SIDE_ENCRYPTION")?;
         let endpoint = env::var("SCCACHE_ENDPOINT").ok();
-        let key_prefix = env::var("SCCACHE_S3_KEY_PREFIX")
-            .ok()
-            .as_ref()
-            .map(|s| s.trim_end_matches('/'))
-            .filter(|s| !s.is_empty())
-            .map(|s| s.to_owned() + "/")
-            .unwrap_or_default();
+        let key_prefix = key_prefix_from_env_var("SCCACHE_S3_KEY_PREFIX");
+        let enable_virtual_host_style = bool_from_env_var("SCCACHE_S3_ENABLE_VIRTUAL_HOST_STYLE")?;
 
         Some(S3CacheConfig {
             bucket,
@@ -554,6 +666,8 @@ fn config_from_env() -> Result<EnvConfig> {
             key_prefix,
             endpoint,
             use_ssl,
+            server_side_encryption,
+            enable_virtual_host_style,
         })
     } else {
         None
@@ -567,21 +681,75 @@ fn config_from_env() -> Result<EnvConfig> {
     }
 
     // ======= redis =======
-    let redis = env::var("SCCACHE_REDIS")
-        .ok()
-        .map(|url| RedisCacheConfig { url });
+    let redis = match (
+        env::var("SCCACHE_REDIS").ok(),
+        env::var("SCCACHE_REDIS_ENDPOINT").ok(),
+        env::var("SCCACHE_REDIS_CLUSTER_ENDPOINTS").ok(),
+    ) {
+        (None, None, None) => None,
+        (url, endpoint, cluster_endpoints) => {
+            let db = number_from_env_var("SCCACHE_REDIS_DB")
+                .transpose()?
+                .unwrap_or(DEFAULT_REDIS_DB);
 
-    // ======= memcached =======
-    let expiration = match env::var("SCCACHE_MEMCACHED_EXPIRATION").ok() {
-        None => DEFAULT_MEMCACHED_CACHE_EXPIRATION,
-        Some(v) => v
-            .parse()
-            .map_err(|err| anyhow!("SCCACHE_MEMCACHED_EXPIRATION value is invalid: {err:?}"))?,
+            let username = env::var("SCCACHE_REDIS_USERNAME").ok();
+            let password = env::var("SCCACHE_REDIS_PASSWORD").ok();
+
+            let ttl = number_from_env_var("SCCACHE_REDIS_EXPIRATION")
+                .or_else(|| number_from_env_var("SCCACHE_REDIS_TTL"))
+                .transpose()?
+                .unwrap_or(DEFAULT_REDIS_CACHE_TTL);
+
+            let key_prefix = key_prefix_from_env_var("SCCACHE_REDIS_KEY_PREFIX");
+
+            Some(RedisCacheConfig {
+                url,
+                endpoint,
+                cluster_endpoints,
+                username,
+                password,
+                db,
+                ttl,
+                key_prefix,
+            })
+        }
     };
 
-    let memcached = env::var("SCCACHE_MEMCACHED")
-        .ok()
-        .map(|url| MemcachedCacheConfig { url, expiration });
+    if env::var_os("SCCACHE_REDIS_EXPIRATION").is_some()
+        && env::var_os("SCCACHE_REDIS_TTL").is_some()
+    {
+        bail!("You mustn't set both SCCACHE_REDIS_EXPIRATION and SCCACHE_REDIS_TTL. Use only one.");
+    }
+
+    // ======= memcached =======
+    let memcached = if let Ok(url) =
+        env::var("SCCACHE_MEMCACHED").or_else(|_| env::var("SCCACHE_MEMCACHED_ENDPOINT"))
+    {
+        let username = env::var("SCCACHE_MEMCACHED_USERNAME").ok();
+        let password = env::var("SCCACHE_MEMCACHED_PASSWORD").ok();
+
+        let expiration = number_from_env_var("SCCACHE_MEMCACHED_EXPIRATION")
+            .transpose()?
+            .unwrap_or(DEFAULT_MEMCACHED_CACHE_EXPIRATION);
+
+        let key_prefix = key_prefix_from_env_var("SCCACHE_MEMCACHED_KEY_PREFIX");
+
+        Some(MemcachedCacheConfig {
+            url,
+            username,
+            password,
+            expiration,
+            key_prefix,
+        })
+    } else {
+        None
+    };
+
+    if env::var_os("SCCACHE_MEMCACHED").is_some()
+        && env::var_os("SCCACHE_MEMCACHED_ENDPOINT").is_some()
+    {
+        bail!("You mustn't set both SCCACHE_MEMCACHED and SCCACHE_MEMCACHED_ENDPOINT. Please, use only SCCACHE_MEMCACHED_ENDPOINT.");
+    }
 
     // ======= GCP/GCS =======
     if (env::var("SCCACHE_GCS_CREDENTIALS_URL").is_ok()
@@ -595,19 +763,11 @@ fn config_from_env() -> Result<EnvConfig> {
     }
 
     let gcs = env::var("SCCACHE_GCS_BUCKET").ok().map(|bucket| {
-        let key_prefix = env::var("SCCACHE_GCS_KEY_PREFIX")
-            .ok()
-            .as_ref()
-            .map(|s| s.trim_end_matches('/'))
-            .filter(|s| !s.is_empty())
-            .unwrap_or_default()
-            .to_owned();
-
-
+        let key_prefix = key_prefix_from_env_var("SCCACHE_GCS_KEY_PREFIX");
 
         if env::var("SCCACHE_GCS_OAUTH_URL").is_ok() {
             eprintln!("SCCACHE_GCS_OAUTH_URL has been deprecated");
-            eprintln!("if you intend to use vm metadata for auth, please set correct service account intead");
+            eprintln!("if you intend to use vm metadata for auth, please set correct service account instead");
         }
 
         let credential_url = env::var("SCCACHE_GCS_CREDENTIALS_URL").ok();
@@ -616,17 +776,17 @@ fn config_from_env() -> Result<EnvConfig> {
         let service_account = env::var("SCCACHE_GCS_SERVICE_ACCOUNT").ok();
 
         let rw_mode = match env::var("SCCACHE_GCS_RW_MODE").as_ref().map(String::as_str) {
-            Ok("READ_ONLY") => GCSCacheRWMode::ReadOnly,
-            Ok("READ_WRITE") => GCSCacheRWMode::ReadWrite,
+            Ok("READ_ONLY") => CacheModeConfig::ReadOnly,
+            Ok("READ_WRITE") => CacheModeConfig::ReadWrite,
             // TODO: unsure if these should warn during the configuration loading
             // or at the time when they're actually used to connect to GCS
             Ok(_) => {
-                warn!("Invalid SCCACHE_GCS_RW_MODE-- defaulting to READ_ONLY.");
-                GCSCacheRWMode::ReadOnly
+                warn!("Invalid SCCACHE_GCS_RW_MODE -- defaulting to READ_ONLY.");
+                CacheModeConfig::ReadOnly
             }
             _ => {
-                warn!("No SCCACHE_GCS_RW_MODE specified-- defaulting to READ_ONLY.");
-                GCSCacheRWMode::ReadOnly
+                warn!("No SCCACHE_GCS_RW_MODE specified -- defaulting to READ_ONLY.");
+                CacheModeConfig::ReadOnly
             }
         };
 
@@ -648,17 +808,13 @@ fn config_from_env() -> Result<EnvConfig> {
             enabled: true,
             version,
         })
-    } else if let Ok(enabled) = env::var("SCCACHE_GHA_ENABLED") {
-        // If only SCCACHE_GHA_ENABLED has been set, enable with
+    } else if bool_from_env_var("SCCACHE_GHA_ENABLED")?.unwrap_or(false) {
+        // If only SCCACHE_GHA_ENABLED has been set to the true value, enable with
         // default version.
-        if enabled == "on" || enabled == "true" {
-            Some(GHACacheConfig {
-                enabled: true,
-                version: "".to_string(),
-            })
-        } else {
-            None
-        }
+        Some(GHACacheConfig {
+            enabled: true,
+            version: "".to_string(),
+        })
     } else {
         None
     };
@@ -668,13 +824,7 @@ fn config_from_env() -> Result<EnvConfig> {
         env::var("SCCACHE_AZURE_CONNECTION_STRING"),
         env::var("SCCACHE_AZURE_BLOB_CONTAINER"),
     ) {
-        let key_prefix = env::var("SCCACHE_AZURE_KEY_PREFIX")
-            .ok()
-            .as_ref()
-            .map(|s| s.trim_end_matches('/'))
-            .filter(|s| !s.is_empty())
-            .unwrap_or_default()
-            .to_owned();
+        let key_prefix = key_prefix_from_env_var("SCCACHE_AZURE_KEY_PREFIX");
         Some(AzureCacheConfig {
             connection_string,
             container,
@@ -686,13 +836,7 @@ fn config_from_env() -> Result<EnvConfig> {
 
     // ======= WebDAV =======
     let webdav = if let Ok(endpoint) = env::var("SCCACHE_WEBDAV_ENDPOINT") {
-        let key_prefix = env::var("SCCACHE_WEBDAV_KEY_PREFIX")
-            .ok()
-            .as_ref()
-            .map(|s| s.trim_end_matches('/'))
-            .filter(|s| !s.is_empty())
-            .unwrap_or_default()
-            .to_owned();
+        let key_prefix = key_prefix_from_env_var("SCCACHE_WEBDAV_KEY_PREFIX");
         let username = env::var("SCCACHE_WEBDAV_USERNAME").ok();
         let password = env::var("SCCACHE_WEBDAV_PASSWORD").ok();
         let token = env::var("SCCACHE_WEBDAV_TOKEN").ok();
@@ -708,16 +852,70 @@ fn config_from_env() -> Result<EnvConfig> {
         None
     };
 
+    // ======= OSS =======
+    let oss = if let Ok(bucket) = env::var("SCCACHE_OSS_BUCKET") {
+        let endpoint = env::var("SCCACHE_OSS_ENDPOINT").ok();
+        let key_prefix = key_prefix_from_env_var("SCCACHE_OSS_KEY_PREFIX");
+
+        let no_credentials = bool_from_env_var("SCCACHE_OSS_NO_CREDENTIALS")?.unwrap_or(false);
+
+        Some(OSSCacheConfig {
+            bucket,
+            endpoint,
+            key_prefix,
+            no_credentials,
+        })
+    } else {
+        None
+    };
+
+    if oss
+        .as_ref()
+        .map(|oss| oss.no_credentials)
+        .unwrap_or_default()
+        && (env::var_os("ALIBABA_CLOUD_ACCESS_KEY_ID").is_some()
+            || env::var_os("ALIBABA_CLOUD_ACCESS_KEY_SECRET").is_some())
+    {
+        bail!("If setting OSS credentials, SCCACHE_OSS_NO_CREDENTIALS must not be set.");
+    }
+
     // ======= Local =======
     let disk_dir = env::var_os("SCCACHE_DIR").map(PathBuf::from);
     let disk_sz = env::var("SCCACHE_CACHE_SIZE")
         .ok()
         .and_then(|v| parse_size(&v));
 
-    let disk = if disk_dir.is_some() || disk_sz.is_some() {
+    let mut preprocessor_mode_config = PreprocessorCacheModeConfig::activated();
+    let preprocessor_mode_overridden = if let Some(value) = bool_from_env_var("SCCACHE_DIRECT")? {
+        preprocessor_mode_config.use_preprocessor_cache_mode = value;
+        true
+    } else {
+        false
+    };
+
+    let (disk_rw_mode, disk_rw_mode_overridden) = match env::var("SCCACHE_LOCAL_RW_MODE")
+        .as_ref()
+        .map(String::as_str)
+    {
+        Ok("READ_ONLY") => (CacheModeConfig::ReadOnly, true),
+        Ok("READ_WRITE") => (CacheModeConfig::ReadWrite, true),
+        Ok(_) => {
+            warn!("Invalid SCCACHE_LOCAL_RW_MODE -- defaulting to READ_WRITE.");
+            (CacheModeConfig::ReadWrite, false)
+        }
+        _ => (CacheModeConfig::ReadWrite, false),
+    };
+
+    let any_overridden = disk_dir.is_some()
+        || disk_sz.is_some()
+        || preprocessor_mode_overridden
+        || disk_rw_mode_overridden;
+    let disk = if any_overridden {
         Some(DiskCacheConfig {
             dir: disk_dir.unwrap_or_else(default_disk_cache_dir),
             size: disk_sz.unwrap_or_else(default_disk_cache_size),
+            preprocessor_cache_mode: preprocessor_mode_config,
+            rw_mode: disk_rw_mode,
         })
     } else {
         None
@@ -757,6 +955,7 @@ fn config_from_env() -> Result<EnvConfig> {
         redis,
         s3,
         webdav,
+        oss,
         concdisk,
     };
 
@@ -1011,7 +1210,7 @@ pub mod server {
             .collect()
     }
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Debug, Serialize, Deserialize, PartialEq)]
     #[serde(tag = "type")]
     #[serde(deny_unknown_fields)]
     pub enum BuilderType {
@@ -1035,7 +1234,7 @@ pub mod server {
         },
     }
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Debug, Serialize, Deserialize, PartialEq)]
     #[serde(tag = "type")]
     #[serde(deny_unknown_fields)]
     pub enum SchedulerAuth {
@@ -1047,12 +1246,13 @@ pub mod server {
         Token { token: String },
     }
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Debug, Serialize, Deserialize, PartialEq)]
     #[serde(deny_unknown_fields)]
     pub struct Config {
         pub builder: BuilderType,
         pub cache_dir: PathBuf,
         pub public_addr: SocketAddr,
+        pub bind_address: Option<SocketAddr>,
         pub scheduler_url: HTTPUrl,
         pub scheduler_auth: SchedulerAuth,
         #[serde(default = "default_toolchain_cache_size")]
@@ -1067,7 +1267,8 @@ pub mod server {
 #[test]
 fn test_parse_size() {
     assert_eq!(None, parse_size(""));
-    assert_eq!(None, parse_size("100"));
+    assert_eq!(None, parse_size("bogus value"));
+    assert_eq!(Some(100), parse_size("100"));
     assert_eq!(Some(2048), parse_size("2K"));
     assert_eq!(Some(10 * 1024 * 1024), parse_size("10M"));
     assert_eq!(Some(TEN_GIGS), parse_size("10G"));
@@ -1086,9 +1287,17 @@ fn config_overrides() {
             disk: Some(DiskCacheConfig {
                 dir: "/env-cache".into(),
                 size: 5,
+                preprocessor_cache_mode: Default::default(),
+                rw_mode: CacheModeConfig::ReadWrite,
             }),
             redis: Some(RedisCacheConfig {
-                url: "myotherredisurl".to_owned(),
+                endpoint: Some("myotherredisurl".to_owned()),
+                ttl: 24 * 3600,
+                key_prefix: "/redis/prefix".into(),
+                db: 10,
+                username: Some("user".to_owned()),
+                password: Some("secret".to_owned()),
+                ..Default::default()
             }),
             ..Default::default()
         },
@@ -1099,13 +1308,20 @@ fn config_overrides() {
             disk: Some(DiskCacheConfig {
                 dir: "/file-cache".into(),
                 size: 15,
+                preprocessor_cache_mode: Default::default(),
+                rw_mode: CacheModeConfig::ReadWrite,
             }),
             memcached: Some(MemcachedCacheConfig {
                 url: "memurl".to_owned(),
                 expiration: 24 * 3600,
+                key_prefix: String::new(),
+                ..Default::default()
             }),
             redis: Some(RedisCacheConfig {
-                url: "myredisurl".to_owned(),
+                url: Some("myredisurl".to_owned()),
+                ttl: 25 * 3600,
+                key_prefix: String::new(),
+                ..Default::default()
             }),
             ..Default::default()
         },
@@ -1117,11 +1333,19 @@ fn config_overrides() {
         Config::from_env_and_file_configs(env_conf, file_conf),
         Config {
             cache: Some(CacheType::Redis(RedisCacheConfig {
-                url: "myotherredisurl".to_owned()
+                endpoint: Some("myotherredisurl".to_owned()),
+                ttl: 24 * 3600,
+                key_prefix: "/redis/prefix".into(),
+                db: 10,
+                username: Some("user".to_owned()),
+                password: Some("secret".to_owned()),
+                ..Default::default()
             }),),
             fallback_cache: DiskCacheConfig {
                 dir: "/env-cache".into(),
                 size: 5,
+                preprocessor_cache_mode: Default::default(),
+                rw_mode: CacheModeConfig::ReadWrite,
             },
             dist: Default::default(),
             server_startup_timeout: None,
@@ -1131,22 +1355,25 @@ fn config_overrides() {
 
 #[test]
 #[serial]
+#[cfg(feature = "s3")]
 fn test_s3_no_credentials_conflict() {
     env::set_var("SCCACHE_S3_NO_CREDENTIALS", "true");
     env::set_var("SCCACHE_BUCKET", "my-bucket");
     env::set_var("AWS_ACCESS_KEY_ID", "aws-access-key-id");
     env::set_var("AWS_SECRET_ACCESS_KEY", "aws-secret-access-key");
 
-    let error = config_from_env().unwrap_err();
-    assert_eq!(
-        "If setting S3 credentials, SCCACHE_S3_NO_CREDENTIALS must not be set.",
-        error.to_string()
-    );
+    let cfg = config_from_env();
 
     env::remove_var("SCCACHE_S3_NO_CREDENTIALS");
     env::remove_var("SCCACHE_BUCKET");
     env::remove_var("AWS_ACCESS_KEY_ID");
     env::remove_var("AWS_SECRET_ACCESS_KEY");
+
+    let error = cfg.unwrap_err();
+    assert_eq!(
+        "If setting S3 credentials, SCCACHE_S3_NO_CREDENTIALS must not be set.",
+        error.to_string()
+    );
 }
 
 #[test]
@@ -1155,14 +1382,16 @@ fn test_s3_no_credentials_invalid() {
     env::set_var("SCCACHE_S3_NO_CREDENTIALS", "yes");
     env::set_var("SCCACHE_BUCKET", "my-bucket");
 
-    let error = config_from_env().unwrap_err();
-    assert_eq!(
-        "SCCACHE_S3_NO_CREDENTIALS must be 'true', '1', 'false', or '0'.",
-        error.to_string()
-    );
+    let cfg = config_from_env();
 
     env::remove_var("SCCACHE_S3_NO_CREDENTIALS");
     env::remove_var("SCCACHE_BUCKET");
+
+    let error = cfg.unwrap_err();
+    assert_eq!(
+        "SCCACHE_S3_NO_CREDENTIALS must be 'true', 'on', '1', 'false', 'off' or '0'.",
+        error.to_string()
+    );
 }
 
 #[test]
@@ -1171,7 +1400,12 @@ fn test_s3_no_credentials_valid_true() {
     env::set_var("SCCACHE_S3_NO_CREDENTIALS", "true");
     env::set_var("SCCACHE_BUCKET", "my-bucket");
 
-    let env_cfg = config_from_env().unwrap();
+    let cfg = config_from_env();
+
+    env::remove_var("SCCACHE_S3_NO_CREDENTIALS");
+    env::remove_var("SCCACHE_BUCKET");
+
+    let env_cfg = cfg.unwrap();
     match env_cfg.cache.s3 {
         Some(S3CacheConfig {
             ref bucket,
@@ -1183,9 +1417,6 @@ fn test_s3_no_credentials_valid_true() {
         }
         None => unreachable!(),
     };
-
-    env::remove_var("SCCACHE_S3_NO_CREDENTIALS");
-    env::remove_var("SCCACHE_BUCKET");
 }
 
 #[test]
@@ -1194,7 +1425,12 @@ fn test_s3_no_credentials_valid_false() {
     env::set_var("SCCACHE_S3_NO_CREDENTIALS", "false");
     env::set_var("SCCACHE_BUCKET", "my-bucket");
 
-    let env_cfg = config_from_env().unwrap();
+    let cfg = config_from_env();
+
+    env::remove_var("SCCACHE_S3_NO_CREDENTIALS");
+    env::remove_var("SCCACHE_BUCKET");
+
+    let env_cfg = cfg.unwrap();
     match env_cfg.cache.s3 {
         Some(S3CacheConfig {
             ref bucket,
@@ -1206,18 +1442,23 @@ fn test_s3_no_credentials_valid_false() {
         }
         None => unreachable!(),
     };
-
-    env::remove_var("SCCACHE_S3_NO_CREDENTIALS");
-    env::remove_var("SCCACHE_BUCKET");
 }
 
 #[test]
+#[serial]
+#[cfg(feature = "gcs")]
 fn test_gcs_service_account() {
     env::set_var("SCCACHE_GCS_BUCKET", "my-bucket");
     env::set_var("SCCACHE_GCS_SERVICE_ACCOUNT", "my@example.com");
     env::set_var("SCCACHE_GCS_RW_MODE", "READ_WRITE");
 
-    let env_cfg = config_from_env().unwrap();
+    let cfg = config_from_env();
+
+    env::remove_var("SCCACHE_GCS_BUCKET");
+    env::remove_var("SCCACHE_GCS_SERVICE_ACCOUNT");
+    env::remove_var("SCCACHE_GCS_RW_MODE");
+
+    let env_cfg = cfg.unwrap();
     match env_cfg.cache.gcs {
         Some(GCSCacheConfig {
             ref bucket,
@@ -1227,14 +1468,10 @@ fn test_gcs_service_account() {
         }) => {
             assert_eq!(bucket, "my-bucket");
             assert_eq!(service_account, Some("my@example.com".to_string()));
-            assert_eq!(rw_mode, GCSCacheRWMode::ReadWrite);
+            assert_eq!(rw_mode, CacheModeConfig::ReadWrite);
         }
         None => unreachable!(),
     };
-
-    env::remove_var("SCCACHE_GCS_BUCKET");
-    env::remove_var("SCCACHE_GCS_SERVICE_ACCOUNT");
-    env::remove_var("SCCACHE_GCS_RW_MODE");
 }
 
 #[test]
@@ -1276,11 +1513,24 @@ enabled = true
 version = "sccache"
 
 [cache.memcached]
-url = "..."
-expiration = 86400
+# Deprecated alias for `endpoint`
+# url = "127.0.0.1:11211"
+endpoint = "tcp://127.0.0.1:11211"
+# Username and password for authentication
+username = "user"
+password = "passwd"
+expiration = 90000
+key_prefix = "/custom/prefix/if/need"
 
 [cache.redis]
-url = "redis://user:passwd@1.2.3.4:6379/1"
+url = "redis://user:passwd@1.2.3.4:6379/?db=1"
+endpoint = "redis://127.0.0.1:6379"
+cluster_endpoints = "tcp://10.0.0.1:6379,redis://10.0.0.2:6379"
+username = "another_user"
+password = "new_passwd"
+db = 12
+expiration = 86400
+key_prefix = "/my/redis/cache"
 
 [cache.s3]
 bucket = "name"
@@ -1289,6 +1539,7 @@ endpoint = "s3-us-east-1.amazonaws.com"
 use_ssl = true
 key_prefix = "s3prefix"
 no_credentials = true
+server_side_encryption = false
 
 [cache.webdav]
 endpoint = "http://127.0.0.1:8080"
@@ -1297,6 +1548,12 @@ username = "webdavusername"
 password = "webdavpassword"
 token = "webdavtoken"
 
+[cache.oss]
+bucket = "name"
+endpoint = "oss-us-east-1.aliyuncs.com"
+key_prefix = "ossprefix"
+no_credentials = true
+
 [cache.concdisk]
 dir = "/tmp/.cache/sccache_concurrent"
 size = 7516192768 # 7 GiBytes
@@ -1312,12 +1569,14 @@ durable_fs = false
                 disk: Some(DiskCacheConfig {
                     dir: PathBuf::from("/tmp/.cache/sccache"),
                     size: 7 * 1024 * 1024 * 1024,
+                    preprocessor_cache_mode: PreprocessorCacheModeConfig::activated(),
+                    rw_mode: CacheModeConfig::ReadWrite,
                 }),
                 gcs: Some(GCSCacheConfig {
                     bucket: "bucket".to_owned(),
                     cred_path: Some("/psst/secret/cred".to_string()),
                     service_account: Some("example_service_account".to_string()),
-                    rw_mode: GCSCacheRWMode::ReadOnly,
+                    rw_mode: CacheModeConfig::ReadOnly,
                     key_prefix: "prefix".into(),
                     credential_url: None,
                 }),
@@ -1326,11 +1585,21 @@ durable_fs = false
                     version: "sccache".to_string()
                 }),
                 redis: Some(RedisCacheConfig {
-                    url: "redis://user:passwd@1.2.3.4:6379/1".to_owned(),
+                    url: Some("redis://user:passwd@1.2.3.4:6379/?db=1".to_owned()),
+                    endpoint: Some("redis://127.0.0.1:6379".to_owned()),
+                    cluster_endpoints: Some("tcp://10.0.0.1:6379,redis://10.0.0.2:6379".to_owned()),
+                    username: Some("another_user".to_owned()),
+                    password: Some("new_passwd".to_owned()),
+                    db: 12,
+                    ttl: 24 * 3600,
+                    key_prefix: "/my/redis/cache".into(),
                 }),
                 memcached: Some(MemcachedCacheConfig {
-                    url: "...".to_owned(),
-                    expiration: 24 * 3600
+                    url: "tcp://127.0.0.1:11211".to_owned(),
+                    username: Some("user".to_owned()),
+                    password: Some("passwd".to_owned()),
+                    expiration: 25 * 3600,
+                    key_prefix: "/custom/prefix/if/need".into(),
                 }),
                 s3: Some(S3CacheConfig {
                     bucket: "name".to_owned(),
@@ -1339,6 +1608,8 @@ durable_fs = false
                     use_ssl: Some(true),
                     key_prefix: "s3prefix".into(),
                     no_credentials: true,
+                    server_side_encryption: Some(false),
+                    enable_virtual_host_style: None,
                 }),
                 webdav: Some(WebdavCacheConfig {
                     endpoint: "http://127.0.0.1:8080".to_string(),
@@ -1347,6 +1618,12 @@ durable_fs = false
                     password: Some("webdavpassword".to_string()),
                     token: Some("webdavtoken".to_string()),
                 }),
+                oss: Some(OSSCacheConfig {
+                    bucket: "name".to_owned(),
+                    endpoint: Some("oss-us-east-1.aliyuncs.com".to_owned()),
+                    key_prefix: "ossprefix".into(),
+                    no_credentials: true,
+                }),
                 concdisk: Some(ConcurrentDiskCacheConfig {
                     dir: PathBuf::from("/tmp/.cache/sccache_concurrent"),
                     size: 7 * 1024 * 1024 * 1024,
@@ -1374,3 +1651,65 @@ durable_fs = false
         }
     )
 }
+
+#[test]
+#[cfg(feature = "dist-server")]
+fn server_toml_parse() {
+    use server::BuilderType;
+    use server::SchedulerAuth;
+    const CONFIG_STR: &str = r#"
+    # This is where client toolchains will be stored.
+    cache_dir = "/tmp/toolchains"
+    # The maximum size of the toolchain cache, in bytes.
+    # If unspecified the default is 10GB.
+    toolchain_cache_size = 10737418240
+    # A public IP address and port that clients will use to connect to this builder.
+    public_addr = "192.168.1.1:10501"
+    # The socket address the builder will listen on.
+    bind_address = "0.0.0.0:10501"
+    # The URL used to connect to the scheduler (should use https, given an ideal
+    # setup of a HTTPS server in front of the scheduler)
+    scheduler_url = "https://192.168.1.1"
+
+    [builder]
+    type = "overlay"
+    # The directory under which a sandboxed filesystem will be created for builds.
+    build_dir = "/tmp/build"
+    # The path to the bubblewrap version 0.3.0+ `bwrap` binary.
+    bwrap_path = "/usr/bin/bwrap"
+
+    [scheduler_auth]
+    type = "jwt_token"
+    # This will be generated by the `generate-jwt-hs256-server-token` command or
+    # provided by an administrator of the sccache cluster.
+    token = "my server's token"
+    "#;
+
+    let server_config: server::Config = toml::from_str(CONFIG_STR).expect("Is valid toml.");
+    assert_eq!(
+        server_config,
+        server::Config {
+            builder: BuilderType::Overlay {
+                build_dir: PathBuf::from("/tmp/build"),
+                bwrap_path: PathBuf::from("/usr/bin/bwrap"),
+            },
+            cache_dir: PathBuf::from("/tmp/toolchains"),
+            public_addr: "192.168.1.1:10501"
+                .parse()
+                .expect("Public address must be valid socket address"),
+            bind_address: Some(
+                "0.0.0.0:10501"
+                    .parse()
+                    .expect("Bind address must be valid socket address")
+            ),
+
+            scheduler_url: parse_http_url("https://192.168.1.1")
+                .map(|url| { HTTPUrl::from_url(url) })
+                .expect("Scheduler url must be valid url str"),
+            scheduler_auth: SchedulerAuth::JwtToken {
+                token: "my server's token".to_owned()
+            },
+            toolchain_cache_size: 10737418240,
+        }
+    )
+}
