@@ -43,27 +43,33 @@ const TC_CACHE_SIZE: u64 = 1024 * 1024 * 1024; // 1 gig
 pub fn start_local_daemon(cfg_path: &Path, cached_cfg_path: &Path) {
     // Don't run this with run() because on Windows `wait_with_output`
     // will hang because the internal server process is not detached.
-    let _ = sccache_command()
+    if !sccache_command()
         .arg("--start-server")
         // Uncomment following lines to debug locally.
-        .env("SCCACHE_LOG", "debug")
-        .env("SCCACHE_ERROR_LOG", "/tmp/sccache_local_daemon.txt")
+        // .env("SCCACHE_LOG", "sccache=trace")
+        // .env("RUST_LOG_STYLE", "never")
+        // .env(
+        //     "SCCACHE_ERROR_LOG",
+        //     env::temp_dir().join("sccache_local_daemon.txt"),
+        // )
         .env("SCCACHE_CONF", cfg_path)
         .env("SCCACHE_CACHED_CONF", cached_cfg_path)
         .status()
         .unwrap()
-        .success();
+        .success()
+    {
+        panic!("Failed to start local daemon");
+    }
 }
 
-pub fn stop_local_daemon() {
+pub fn stop_local_daemon() -> bool {
     trace!("sccache --stop-server");
-    drop(
-        sccache_command()
-            .arg("--stop-server")
-            .stdout(Stdio::null())
-            .stderr(Stdio::null())
-            .status(),
-    );
+    sccache_command()
+        .arg("--stop-server")
+        .stdout(Stdio::null())
+        .stderr(Stdio::null())
+        .status()
+        .is_ok_and(|status| status.success())
 }
 
 pub fn get_stats<F: 'static + Fn(ServerInfo)>(f: F) {
@@ -104,11 +110,17 @@ pub fn write_source(path: &Path, filename: &str, contents: &str) {
     f.write_all(contents.as_bytes()).unwrap();
 }
 
+pub fn init_cargo(path: &Path, cargo_name: &str) -> PathBuf {
+    let cargo_path = path.join(cargo_name);
+    let source_path = "src";
+    fs::create_dir_all(cargo_path.join(source_path)).unwrap();
+    cargo_path
+}
+
 // Prune any environment variables that could adversely affect test execution.
-pub fn sccache_command() -> Command {
+pub fn prune_command(mut cmd: Command) -> Command {
     use sccache::util::OsStrExt;
 
-    let mut cmd = Command::new(assert_cmd::cargo::cargo_bin("sccache"));
     for (var, _) in env::vars_os() {
         if var.starts_with("SCCACHE_") {
             cmd.env_remove(var);
@@ -117,12 +129,23 @@ pub fn sccache_command() -> Command {
     cmd
 }
 
+pub fn sccache_command() -> Command {
+    prune_command(Command::new(assert_cmd::cargo::cargo_bin("sccache")))
+}
+
+pub fn cargo_command() -> Command {
+    prune_command(Command::new("cargo"))
+}
+
 #[cfg(feature = "dist-server")]
 pub fn sccache_dist_path() -> PathBuf {
     assert_cmd::cargo::cargo_bin("sccache-dist")
 }
 
-pub fn sccache_client_cfg(tmpdir: &Path) -> sccache::config::FileConfig {
+pub fn sccache_client_cfg(
+    tmpdir: &Path,
+    preprocessor_cache_mode: bool,
+) -> sccache::config::FileConfig {
     let cache_relpath = "client-cache";
     let dist_cache_relpath = "client-dist-cache";
     fs::create_dir(tmpdir.join(cache_relpath)).unwrap();
@@ -130,6 +153,10 @@ pub fn sccache_client_cfg(tmpdir: &Path) -> sccache::config::FileConfig {
 
     let disk_cache = sccache::config::DiskCacheConfig {
         dir: tmpdir.join(cache_relpath),
+        preprocessor_cache_mode: sccache::config::PreprocessorCacheModeConfig {
+            use_preprocessor_cache_mode: preprocessor_cache_mode,
+            ..Default::default()
+        },
         ..Default::default()
     };
     sccache::config::FileConfig {
@@ -142,6 +169,7 @@ pub fn sccache_client_cfg(tmpdir: &Path) -> sccache::config::FileConfig {
             redis: None,
             s3: None,
             webdav: None,
+            oss: None,
             concdisk: None,
         },
         dist: sccache::config::DistConfig {
@@ -155,6 +183,7 @@ pub fn sccache_client_cfg(tmpdir: &Path) -> sccache::config::FileConfig {
         server_startup_timeout_ms: None,
     }
 }
+
 #[cfg(feature = "dist-server")]
 fn sccache_scheduler_cfg() -> sccache::config::scheduler::Config {
     sccache::config::scheduler::Config {
@@ -165,6 +194,7 @@ fn sccache_scheduler_cfg() -> sccache::config::scheduler::Config {
         },
     }
 }
+
 #[cfg(feature = "dist-server")]
 fn sccache_server_cfg(
     tmpdir: &Path,
@@ -181,6 +211,7 @@ fn sccache_server_cfg(
         },
         cache_dir: Path::new(CONFIGS_CONTAINER_PATH).join(relpath),
         public_addr: SocketAddr::new(server_ip, SERVER_PORT),
+        bind_address: Some(SocketAddr::from(([0, 0, 0, 0], SERVER_PORT))),
         scheduler_url,
         scheduler_auth: sccache::config::server::SchedulerAuth::Token {
             token: DIST_SERVER_TOKEN.to_owned(),
@@ -217,7 +248,7 @@ impl DistSystem {
     pub fn new(sccache_dist: &Path, tmpdir: &Path) -> Self {
         // Make sure the docker image is available, building it if necessary
         let mut child = Command::new("docker")
-            .args(&["build", "-q", "-t", DIST_IMAGE, "-"])
+            .args(["build", "-q", "-t", DIST_IMAGE, "-"])
             .stdin(Stdio::piped())
             .stdout(Stdio::piped())
             .stderr(Stdio::piped())
@@ -251,7 +282,7 @@ impl DistSystem {
         let scheduler_cfg_container_path =
             Path::new(CONFIGS_CONTAINER_PATH).join(scheduler_cfg_relpath);
         let scheduler_cfg = sccache_scheduler_cfg();
-        fs::File::create(&scheduler_cfg_path)
+        fs::File::create(scheduler_cfg_path)
             .unwrap()
             .write_all(&serde_json::to_vec(&scheduler_cfg).unwrap())
             .unwrap();
@@ -259,7 +290,7 @@ impl DistSystem {
         // Create the scheduler
         let scheduler_name = make_container_name("scheduler");
         let output = Command::new("docker")
-            .args(&[
+            .args([
                 "run",
                 "--name",
                 &scheduler_name,
@@ -272,10 +303,10 @@ impl DistSystem {
                 "--network",
                 "host",
                 "-v",
-                &format!("{}:/sccache-dist", self.sccache_dist.to_str().unwrap()),
+                &format!("{}:/sccache-dist:z", self.sccache_dist.to_str().unwrap()),
                 "-v",
                 &format!(
-                    "{}:{}",
+                    "{}:{}:z",
                     self.tmpdir.to_str().unwrap(),
                     CONFIGS_CONTAINER_PATH
                 ),
@@ -327,7 +358,7 @@ impl DistSystem {
 
         let server_name = make_container_name("server");
         let output = Command::new("docker")
-            .args(&[
+            .args([
                 "run",
                 // Important for the bubblewrap builder
                 "--privileged",
@@ -340,10 +371,10 @@ impl DistSystem {
                 "--network",
                 "host",
                 "-v",
-                &format!("{}:/sccache-dist", self.sccache_dist.to_str().unwrap()),
+                &format!("{}:/sccache-dist:z", self.sccache_dist.to_str().unwrap()),
                 "-v",
                 &format!(
-                    "{}:{}",
+                    "{}:{}:z",
                     self.tmpdir.to_str().unwrap(),
                     CONFIGS_CONTAINER_PATH
                 ),
@@ -395,9 +426,14 @@ impl DistSystem {
             listener.local_addr().unwrap()
         };
         let token = create_server_token(ServerId::new(server_addr), DIST_SERVER_TOKEN);
-        let server =
-            dist::http::Server::new(server_addr, self.scheduler_url().to_url(), token, handler)
-                .unwrap();
+        let server = dist::http::Server::new(
+            server_addr,
+            Some(SocketAddr::from(([0, 0, 0, 0], server_addr.port()))),
+            self.scheduler_url().to_url(),
+            token,
+            handler,
+        )
+        .unwrap();
         let pid = match unsafe { nix::unistd::fork() }.unwrap() {
             ForkResult::Parent { child } => {
                 self.server_pids.push(child);
@@ -406,7 +442,8 @@ impl DistSystem {
             ForkResult::Child => {
                 env::set_var("SCCACHE_LOG", "sccache=trace");
                 env_logger::try_init().unwrap();
-                void::unreachable(server.start().unwrap())
+                server.start().unwrap();
+                unreachable!();
             }
         };
 
@@ -421,7 +458,7 @@ impl DistSystem {
         match handle {
             ServerHandle::Container { cid, url: _ } => {
                 let output = Command::new("docker")
-                    .args(&["restart", cid])
+                    .args(["restart", cid])
                     .output()
                     .unwrap();
                 check_output(&output);
@@ -504,29 +541,29 @@ impl Drop for DistSystem {
 
         if let Some(scheduler_name) = self.scheduler_name.as_ref() {
             droperr!(Command::new("docker")
-                .args(&["logs", scheduler_name])
+                .args(["logs", scheduler_name])
                 .output()
                 .map(|o| logs.push((scheduler_name, o))));
             droperr!(Command::new("docker")
-                .args(&["kill", scheduler_name])
+                .args(["kill", scheduler_name])
                 .output()
                 .map(|o| outputs.push((scheduler_name, o))));
             droperr!(Command::new("docker")
-                .args(&["rm", "-f", scheduler_name])
+                .args(["rm", "-f", scheduler_name])
                 .output()
                 .map(|o| outputs.push((scheduler_name, o))));
         }
         for server_name in self.server_names.iter() {
             droperr!(Command::new("docker")
-                .args(&["logs", server_name])
+                .args(["logs", server_name])
                 .output()
                 .map(|o| logs.push((server_name, o))));
             droperr!(Command::new("docker")
-                .args(&["kill", server_name])
+                .args(["kill", server_name])
                 .output()
                 .map(|o| outputs.push((server_name, o))));
             droperr!(Command::new("docker")
-                .args(&["rm", "-f", server_name])
+                .args(["rm", "-f", server_name])
                 .output()
                 .map(|o| outputs.push((server_name, o))));
         }
